def MakeKerasPooling(layer):
    """
    Create a Keras-compatible pooling layer operation using SOFIE framework.

    This function takes a dictionary representing a pooling layer and its attributes and
    constructs a Keras-compatible pooling layer operation using the SOFIE framework.
    Pooling layers downsample the input tensor by selecting a representative value from
    a group of neighboring values, either by taking the maximum or the average.

    Parameters:
    layer (dict): A dictionary containing layer information including input, output,
                  layer type (the selection rule), the pool size, padding, strides, and data type.

    Returns:
    ROperator_Pool: A SOFIE framework operator representing the pooling layer operation.
    """
    from ROOT.TMVA.Experimental import SOFIE

    # Extract attributes from layer data
    fLayerDType = layer["layerDType"]
    finput = layer["layerInput"]
    foutput = layer["layerOutput"]
    fLayerType = layer["layerType"]
    fLayerInputName = finput[0]
    fLayerOutputName = foutput[0]
    pool_atrr = SOFIE.RAttributes_Pool()
    attributes = layer["layerAttributes"]
    # Set default values for GlobalAveragePooling2D
    fAttrKernelShape = []
    fKerasPadding = "valid"
    fAttrStrides = []
    if fLayerType != "GlobalAveragePooling2D":
        fAttrKernelShape = attributes["pool_size"]
        fKerasPadding = str(attributes["padding"])
        fAttrStrides = attributes["strides"]

    # Set default values
    fAttrDilations = (1, 1)
    fpads = [0, 0, 0, 0, 0, 0]
    pool_atrr.ceil_mode = 0
    pool_atrr.count_include_pad = 0
    pool_atrr.storage_order = 0

    if fKerasPadding == "valid":
        fAttrAutopad = "VALID"
    elif fKerasPadding == "same":
        fAttrAutopad = "NOTSET"
    else:
        raise RuntimeError(
            "TMVA::SOFIE - RModel Keras Parser doesn't yet support Pooling layer with padding " + fKerasPadding
        )
    pool_atrr.dilations = list(fAttrDilations)
    pool_atrr.strides = list(fAttrStrides)
    pool_atrr.pads = fpads
    pool_atrr.kernel_shape = list(fAttrKernelShape)
    pool_atrr.auto_pad = fAttrAutopad

    # Choose pooling type
    if "Max" in fLayerType:
        PoolMode = SOFIE.PoolOpMode.MaxPool
    elif "AveragePool" in fLayerType:
        PoolMode = SOFIE.PoolOpMode.AveragePool
    elif "GlobalAverage" in fLayerType:
        PoolMode = SOFIE.PoolOpMode.GloabalAveragePool
    else:
        raise RuntimeError(
            "TMVA::SOFIE - Unsupported - Operator poolong does not yet support pooling type " + fLayerType
        )

    # Create operator
    if SOFIE.ConvertStringToType(fLayerDType) == SOFIE.ETensorType.FLOAT:
        op = SOFIE.ROperator_Pool["float"](PoolMode, pool_atrr, fLayerInputName, fLayerOutputName)
        return op
    else:
        raise RuntimeError(
            "TMVA::SOFIE - Unsupported - Operator Pooling does not yet support input type " + fLayerDType
        )



# Author: Vincenzo Eduardo Padulano CERN/UPV 2022

################################################################################
# Copyright (C) 1995-2022, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

r'''
\pythondoc TDirectory::TContext

The functionality offered by TContext can be used in PyROOT with a context manager. Here are a few examples:
\code{.py}
import ROOT
from ROOT import TDirectory

with TDirectory.TContext():
    # Open some file here
    file = ROOT.TFile(...)
    # Retrieve contents from the file
    histo = file.Get("myhisto")

# After the 'with' statement, the current directory is restored to ROOT.gROOT
\endcode
\n
\code{.py}
import ROOT
from ROOT import TDirectory

file1 = ROOT.TFile("file1.root", "recreate")
#...
file2 = ROOT.TFile("file2.root", "recreate")
#...
file3 = ROOT.TFile("file3.root", "recreate")

# Before the 'with' statement, the current directory is file3 (the last file opened)
with TDirectory.TContext(file1):
    # Inside the statement, the current directory is file1
    histo = ROOT.TH1F(...)
    histo.Write()

# After the statement, the current directory is restored to file3
\endcode
\n
\code{.py}
import ROOT
from ROOT import TDirectory

file1 = ROOT.TFile("file1.root")
file2 = ROOT.TFile("file2.root")

with TDirectory.TContext(file1, file2):
    # Manage content in file2
    histo = ROOT.TH1F(...)
    # histo will be written to file2
    histo.Write()

# Current directory is set to 'file1.root'
\endcode

Note that TContext restores the current directory to its status before the 'with'
statement, but does not change the status of any file that has been opened inside
the context (e.g. it does not automatically close the file).

\endpythondoc
'''

from . import pythonization


def _TContextExit(ctxt, exc_type, exc_val, exc_tb):
    """
    Destroy the TContext, mimicking its RAII behaviour.
    Signature and return value are imposed by Python, see
    https://docs.python.org/3/library/stdtypes.html#typecontextmanager.
    """
    ctxt.__destruct__()
    return False


@pythonization("TContext", ns="TDirectory")
def pythonize_tcontext(klass):
    """
    TContext pythonizations:
    - __enter__ and __exit__ methods are defined to make it work as a context manager.
    """

    # The user doesn't need to do anything with the TContext object itself, so
    # we don't provide it in the context manager expression
    klass.__enter__ = lambda ctxt: None
    klass.__exit__ = _TContextExit



# -----------------------------------------------------------------------------
#  Author: Danilo Piparo <Danilo.Piparo@cern.ch> CERN
#  Author: Enric Tejedor <enric.tejedor.saavedra@cern.ch> CERN
# -----------------------------------------------------------------------------

################################################################################
# Copyright (C) 1995-2020, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

import sys

from ROOT._jupyroot.helpers import utils

if "win32" not in sys.platform:
    from ROOT._jupyroot.helpers import cppcompleter

# Check if we are in the IPython shell
try:
    import builtins
except ImportError:
    import __builtin__ as builtins  # Py2
_is_ipython = hasattr(builtins, "__IPYTHON__")

if _is_ipython:
    if "win32" not in sys.platform:
        from IPython import get_ipython

        cppcompleter.load_ipython_extension(get_ipython())
    utils.iPythonize()



# Authors:
# * Harshal Shende  03/2021
# * Jonas Rembser 03/2021

################################################################################
# Copyright (C) 1995-2020, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################


from ._rooabsreal import RooAbsReal
from ._utils import _kwargs_to_roocmdargs, cpp_signature, _pack_cmd_args


class RooAbsPdf(RooAbsReal):
    r"""Some member functions of RooAbsPdf that take a RooCmdArg as argument also support keyword arguments.
    So far, this applies to RooAbsPdf::fitTo, RooAbsPdf::plotOn, RooAbsPdf::generate, RooAbsPdf::paramOn, RooAbsPdf::createCdf,
    RooAbsPdf::generateBinned, RooAbsPdf::prepareMultiGen and RooAbsPdf::createNLL.
    For example, the following code is equivalent in PyROOT:
    \code{.py}
    # Directly passing a RooCmdArg:
    pdf.fitTo(data, ROOT.RooFit.Range("r1"))

    # With keyword arguments:
    pdf.fitTo(data, Range="r1")
    \endcode"""

    __cpp_name__ = 'RooAbsPdf'

    @cpp_signature("RooAbsPdf::fitTo()")
    def fitTo(self, *args, **kwargs):
        r"""The RooAbsPdf::fitTo() function is pythonized with the command argument pythonization.
        The keywords must correspond to the CmdArgs of the function.
        """
        import ROOT

        out = self._fitTo["RooLinkedList const&"](args[0], _pack_cmd_args(*args[1:], **kwargs))
        ROOT.SetOwnership(out, True)
        return out

    @cpp_signature(
        "RooPlot *RooAbsPdf::plotOn(RooPlot* frame,"
        "    const RooCmdArg& arg1={}, const RooCmdArg& arg2={},"
        "    const RooCmdArg& arg3={}, const RooCmdArg& arg4={},"
        "    const RooCmdArg& arg5={}, const RooCmdArg& arg6={},"
        "    const RooCmdArg& arg7={}, const RooCmdArg& arg8={},"
        "    const RooCmdArg& arg9={}, const RooCmdArg& arg10={}"
        ") const;"
    )
    def plotOn(self, *args, **kwargs):
        r"""The RooAbsPdf::plotOn() function is pythonized with the command argument pythonization.
        The keywords must correspond to the CmdArgs of the function.
        """
        # Redefinition of `RooAbsPdf.plotOn` for keyword arguments.
        args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
        return self._plotOn(*args, **kwargs)

    @cpp_signature(
        "RooDataSet *RooAbsPdf::generate(const RooArgSet &whatVars,"
        "    const RooCmdArg& arg1={},const RooCmdArg& arg2={},"
        "    const RooCmdArg& arg3={},const RooCmdArg& arg4={},"
        "    const RooCmdArg& arg5={},const RooCmdArg& arg6={}) ;"
    )
    def generate(self, *args, **kwargs):
        r"""The RooAbsPdf::generate() function is pythonized with the command argument pythonization.
        The keywords must correspond to the CmdArgs of the function.
        """
        # Redefinition of `RooAbsPdf.generate` for keyword arguments.
        args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
        return self._generate(*args, **kwargs)

    @cpp_signature(
        "RooPlot *RooAbsPdf::paramOn(RooPlot* frame,"
        "    const RooCmdArg& arg1={}, const RooCmdArg& arg2={},"
        "    const RooCmdArg& arg3={}, const RooCmdArg& arg4={},"
        "    const RooCmdArg& arg5={}, const RooCmdArg& arg6={},"
        "    const RooCmdArg& arg7={}, const RooCmdArg& arg8={}) ;"
    )
    def paramOn(self, *args, **kwargs):
        r"""The RooAbsPdf::paramOn() function is pythonized with the command argument pythonization.
        The keywords must correspond to the CmdArgs of the function.
        """
        # Redefinition of `RooAbsPdf.paramOn` for keyword arguments.
        args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
        return self._paramOn(*args, **kwargs)

    @cpp_signature("RooAbsPdf::createNLL()")
    def createNLL(self, *args, **kwargs):
        r"""The RooAbsPdf::createNLL() function is pythonized with the command argument pythonization.
        The keywords must correspond to the CmdArgs of the function.
        """
        import ROOT

        out = self._createNLL["RooLinkedList const&"](args[0], _pack_cmd_args(*args[1:], **kwargs))
        ROOT.SetOwnership(out, True)
        return out

    @cpp_signature(
        "RooAbsReal *RooAbsPdf::createCdf(const RooArgSet& iset, const RooCmdArg& arg1, const RooCmdArg& arg2={},"
        "    const RooCmdArg& arg3={}, const RooCmdArg& arg4={},"
        "    const RooCmdArg& arg5={}, const RooCmdArg& arg6={},"
        "    const RooCmdArg& arg7={}, const RooCmdArg& arg8={}) ;"
    )
    def createCdf(self, *args, **kwargs):
        r"""The RooAbsPdf::createCdf() function is pythonized with the command argument pythonization.
        The keywords must correspond to the CmdArgs of the function.
        """
        import ROOT

        args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
        out = self._createCdf(*args, **kwargs)
        ROOT.SetOwnership(out, True)
        return out

    @cpp_signature(
        "RooDataHist *RooAbsPdf::generateBinned(const RooArgSet &whatVars,"
        "   const RooCmdArg& arg1={},const RooCmdArg& arg2={},"
        "    const RooCmdArg& arg3={},const RooCmdArg& arg4={},"
        "    const RooCmdArg& arg5={},const RooCmdArg& arg6={}) const;"
    )
    def generateBinned(self, *args, **kwargs):
        r"""The RooAbsPdf::generateBinned() function is pythonized with the command argument pythonization.
        The keywords must correspond to the CmdArgs of the function.
        """
        import ROOT

        args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
        out = self._generateBinned(*args, **kwargs)
        ROOT.SetOwnership(out, True)
        return out

    @cpp_signature(
        "GenSpec *RooAbsPdf::prepareMultiGen(const RooArgSet &whatVars,"
        "    const RooCmdArg& arg1={},const RooCmdArg& arg2={},"
        "    const RooCmdArg& arg3={},const RooCmdArg& arg4={},"
        "    const RooCmdArg& arg5={},const RooCmdArg& arg6={}) ;"
    )
    def prepareMultiGen(self, *args, **kwargs):
        r"""The RooAbsPdf::prepareMultiGen() function is pythonized with the command argument pythonization.
        The keywords must correspond to the CmdArgs of the function.
        """
        # Redefinition of `RooAbsPdf.prepareMultiGen` for keyword arguments.
        args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
        return self._prepareMultiGen(*args, **kwargs)



#  @author Vincenzo Eduardo Padulano
#  @author Enric Tejedor
#  @date 2021-02

################################################################################
# Copyright (C) 1995-2022, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Callable, Optional

# Type hints only
if TYPE_CHECKING:
    # Avoid circular imports
    from .HeadNode import HeadNode
    from .Operation import Operation

logger = logging.getLogger(__name__)


class Node(object):
    """
    A Class that represents a node in RDataFrame operations graph. A Node
    houses an operation and has references to children nodes.

    Attributes:
        get_head: A function returning the head node in the graph.

        node_id: The id of this node, given sequentially in the order of
            creation with respect to the head node of the graph.

        operation: The operation that this node represents. `None` if the node
            is the head node.

        parent: A reference to the parent node of this node. `None` if the node
            is the head node.

        nchildren: A counter of how many children this node has.

        _new_op_name (str): The name of the new incoming operation of the next
            child, which is the last child node among the current node's
            children.

        value: The computed value after executing the operation in the current
            node for a particular DistRDF graph. This is permanently :obj:`None`
            for transformation nodes and the action nodes get a
            :obj:`ROOT.RResultPtr` after event-loop execution.

        has_user_references (bool): A flag to check whether the node has
            direct user references, that is if it is assigned to a variable.
            Default value is :obj:`True`, turns to :obj:`False` if the proxy
            that wraps the node gets garbage collected by Python.

        rdf_node: A reference to the result of calling a function of the
            RDataFrame API with the current operation. This is practically a
            node of the true computation graph, which is being executed in some
            distributed task. It is a transient attribute. On the client, this
            is filled when the operation being called is a transformation. It is
            done to ensure information regarding e.g. column names and types is
            populated and available locally with the right dependencies. On a
            worker, this attribute can represent any node of the C++ computation
            graph and its created and processed within the worker.
    """

    def __init__(self, get_head: Callable[[], HeadNode], node_id: int = 0,
                 operation: "Operation" = None, parent: Node = None):
        self.get_head = get_head
        self.node_id = node_id
        self.operation = operation
        self.parent = parent
        self.nchildren: int = 0
        self._new_op_name: str = ""
        self.value = None
        self.has_user_references: bool = True
        self.rdf_node = None

        # This is the internal attribute for the 'parent_id' property. It is
        # serialized from the local information then deserialized and set in a
        # distributed task.
        self._parent_id: Optional[int] = parent.node_id if parent is not None else None

    @property
    def parent_id(self) -> Optional[int]:
        """Retrieves the id of the parent node."""
        return self._parent_id

    @parent_id.setter
    def parent_id(self, value: Optional[int]):
        """
        Sets the id of the parent node. Only present to enable setting this
        attribute when deserializing the node in a distributed task.
        """
        self._parent_id = value

    def __getstate__(self):
        """
        Serialize the minimum amount of information needed in a distributed task
        to execute the operation corresponding to this node of the graph.
        """
        return {"operation": self.operation, "parent_id": self.parent_id}

    def __setstate__(self, state):
        self.operation = state["operation"]
        self.parent_id = state["parent_id"]

    def is_prunable(self) -> bool:
        """
        Checks whether the current node can be pruned from the computational
        graph.

        Returns:
            bool: True if the node has no children and no user references or
            its value has already been computed, False otherwise.
        """
        logger.debug(f"Checking node {self.node_id} for pruning")

        if self.nchildren == 0 and (not self.has_user_references or self.value is not None):

            # ***** Condition 1 *****
            # If the node does not have children, it might be prunable.

            # ***** Condition 2 *****
            # If the node is wrapped by a proxy which is not directly
            # assigned to a variable, then it will be flagged for pruning

            # ***** Condition 3 *****
            # If the current node's value was already computed, it should
            # get pruned. Only action nodes may possess a value attribute
            # which is not None

            logger.debug(f"node {self.node_id} can be pruned")

            # Decrement children count of the parent node of this node, so that
            # when the pruning will consider the parent, this child will not be
            # counted.
            self.parent.nchildren -= 1

            return True

        logger.debug(f"node {self.node_id} should not be pruned")

        return False



## \file
## \ingroup tutorial_roofit_main
## \notebook
## Basic functionality: demonstration of various plotting styles of data, functions in a RooPlot
##
## \macro_image
## \macro_code
## \macro_output
##
## \date February 2018
## \authors Clemens Lange, Wouter Verkerke (C++ version)

import ROOT


# Set up model
# ---------------------

# Create observables
x = ROOT.RooRealVar("x", "x", -10, 10)

# Create Gaussian
sigma = ROOT.RooRealVar("sigma", "sigma", 3, 0.1, 10)
mean = ROOT.RooRealVar("mean", "mean", -3, -10, 10)
gauss = ROOT.RooGaussian("gauss", "gauss", x, mean, sigma)

# Generate a sample of 100 events with sigma=3
data = gauss.generate({x}, 100)

# Fit pdf to data
gauss.fitTo(data, PrintLevel=-1)

# Make plot frames
# -------------------------------

# Make four plot frames to demonstrate various plotting features
frame1 = x.frame(Name="xframe", Title="Red Curve / SumW2 Histo errors", Bins=20)
frame2 = x.frame(Name="xframe", Title="Dashed Curve / No XError bars", Bins=20)
frame3 = x.frame(Name="xframe", Title="Filled Curve / Blue Histo", Bins=20)
frame4 = x.frame(Name="xframe", Title="Partial Range / Filled Bar chart", Bins=20)

# Data plotting styles
# ---------------------------------------

# Use sqrt(sum(weights^2)) error instead of Poisson errors
data.plotOn(frame1, DataError="SumW2")

# Remove horizontal error bars
data.plotOn(frame2, XErrorSize=0)

# Blue markers and error bors
data.plotOn(frame3, MarkerColor="b", LineColor="b")

# Filled bar chart
data.plotOn(frame4, DrawOption="B", DataError=None, XErrorSize=0, FillColor="kGray")

# Function plotting styles
# -----------------------------------------------

# Change line color to red
gauss.plotOn(frame1, LineColor="r")

# Change line style to dashed
gauss.plotOn(frame2, LineStyle="--")

# Filled shapes in green color
gauss.plotOn(frame3, MoveToBack=True, DrawOption="F", FillColor="kOrange")

#
gauss.plotOn(frame4, Range=(-8, 3), LineColor="m")

c = ROOT.TCanvas("rf107_plotstyles", "rf107_plotstyles", 800, 800)
c.Divide(2, 2)
c.cd(1)
ROOT.gPad.SetLeftMargin(0.15)
frame1.GetYaxis().SetTitleOffset(1.6)
frame1.Draw()
c.cd(2)
ROOT.gPad.SetLeftMargin(0.15)
frame2.GetYaxis().SetTitleOffset(1.6)
frame2.Draw()
c.cd(3)
ROOT.gPad.SetLeftMargin(0.15)
frame3.GetYaxis().SetTitleOffset(1.6)
frame3.Draw()
c.cd(4)
ROOT.gPad.SetLeftMargin(0.15)
frame4.GetYaxis().SetTitleOffset(1.6)
frame4.Draw()

c.SaveAs("rf107_plotstyles.png")



## \file
## \ingroup tutorial_fit
## \notebook
##  Fitting multiple functions to different ranges of a 1-D histogram
##      Example showing how to fit in a sub-range of an histogram
##  A histogram is created and filled with the bin contents and errors
##  defined in the table below.
##  Three Gaussians are fitted in sub-ranges of this histogram.
##  A new function (a sum of 3 Gaussians) is fitted on another subrange
##  Note that when fitting simple functions, such as Gaussians, the initial
##  values of parameters are automatically computed by ROOT.
##  In the more complicated case of the sum of 3 Gaussians, the initial values
##  of parameters must be given. In this particular case, the initial values
##  are taken from the result of the individual fits.
##
## \macro_image
## \macro_output
## \macro_code
##
## \authors Jonas Rembser, Rene Brun (C++ version)

import ROOT

import numpy as np

n_x = 49

# fmt: off
x = np.array( [ 1.913521, 1.953769, 2.347435, 2.883654, 3.493567, 4.047560,
    4.337210, 4.364347, 4.563004, 5.054247, 5.194183, 5.380521, 5.303213,
    5.384578, 5.563983, 5.728500, 5.685752, 5.080029, 4.251809, 3.372246,
    2.207432, 1.227541, 0.8597788, 0.8220503, 0.8046592, 0.7684097, 0.7469761,
    0.8019787, 0.8362375, 0.8744895, 0.9143721, 0.9462768, 0.9285364,
    0.8954604, 0.8410891, 0.7853871, 0.7100883, 0.6938808, 0.7363682,
    0.7032954, 0.6029015, 0.5600163, 0.7477068, 1.188785, 1.938228, 2.602717,
    3.472962, 4.465014, 5.177035, ], dtype=np.float32,)
# fmt: on

# The histogram are filled with bins defined in the array x.
h = ROOT.TH1F("h", "Example of several fits in subranges", n_x, 85, 134)
h.SetMaximum(7)

for i, x_i in enumerate(x):
    h.SetBinContent(i + 1, x[i])

# Define the parameter array for the total function.
par = np.zeros(9)

# Three TF1 objects are created, one for each subrange.
g1 = ROOT.TF1("g1", "gaus", 85, 95)
g2 = ROOT.TF1("g2", "gaus", 98, 108)
g3 = ROOT.TF1("g3", "gaus", 110, 121)

# The total is the sum of the three, each has three parameters.
total = ROOT.TF1("total", "gaus(0)+gaus(3)+gaus(6)", 85, 125)
total.SetLineColor(2)

# The canvas that the histograms and fit functions are drawn on.
c = ROOT.TCanvas("multifit", "multifit", 800, 400)

# Fit each function and add it to the list of functions. By default, TH1::Fit()
# fits the function on the defined histogram range. You can specify the "R"
# option in the second parameter of TH1::Fit() to restrict the fit to the range
# specified in the TF1 constructor. Alternatively, you can also specify the
# range in the call to TH1::Fit(), which we demonstrate here with the 3rd
# Gaussian. The "+" option needs to be added to the later fits to not replace
# existing fitted functions in the histogram.
h.Fit(g1, "R")
h.Fit(g2, "R+")
h.Fit(g3, "+", "", 110, 121);

# Get the parameters from the fit.
g1.GetParameters(par[:3])
g2.GetParameters(par[3:6])
g3.GetParameters(par[6:])

print(par)

# Use the parameters on the sum.
total.SetParameters(par)

h.Draw()
h.Fit(total, "R+")

# Save the plot for later inspection.
c.SaveAs("multifit.png")



# -*- coding:utf-8 -*-
# -----------------------------------------------------------------------------
#  Authors: Omar Zapata <Omar.Zapata@cern.ch> http://oproject.org
#           Danilo Piparo <Danilo.Piparo@cern.ch> CERN
#           Enric Tejedor enric.tejedor.saavedra@cern.ch> CERN
# -----------------------------------------------------------------------------

################################################################################
# Copyright (C) 1995-2020, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

import importlib
import os
from glob import glob

from ROOT._jupyroot.helpers.handlers import IOHandler, JupyROOTDeclarer, JupyROOTDisplayer, JupyROOTExecutor, Poller

_ioHandler = None
_Poller = None
_Executor = None
_Declarer = None
_Displayer = None


def GetIOHandler():
    global _ioHandler
    if not _ioHandler:
        _ioHandler = IOHandler()
    return _ioHandler


def GetPoller():
    global _Poller
    if not _Poller:
        _Poller = Poller()
        _Poller.start()
    return _Poller


def GetExecutor(poller):
    global _Executor
    if not _Executor:
        _Executor = JupyROOTExecutor(poller)
    return _Executor


def GetDeclarer(poller):
    global _Declarer
    if not _Declarer:
        _Declarer = JupyROOTDeclarer(poller)
    return _Declarer


def GetDisplayer(poller):
    global _Displayer
    if not _Displayer:
        _Displayer = JupyROOTDisplayer(poller)
    return _Displayer


class MagicLoader(object):
    """Class to load JupyROOT Magics"""

    def __init__(self, kernel):
        magics_path = os.path.join(os.path.dirname(__file__), "magics", "*.py")
        for file in glob(magics_path):
            if file != magics_path.replace("*.py", "__init__.py"):
                module_prefix = "ROOT._jupyroot.kernel.magics."
                module_name = os.path.splitext(os.path.basename(file))[0]
                module_path = module_prefix + module_name
                try:
                    module = importlib.import_module(module_path)
                    module.register_magics(kernel)
                except ImportError:
                    raise Exception("Error importing Magic: %s" % module_path)



## \file
## \ingroup tutorial_pdf
## \notebook
## Example macro showing some major probability density functions in ROOT.
## The macro shows four of them with
## respect to their two variables. In order to run the macro type:
##
## ~~~{.cpp}
##   root [0] .x mathcoreStatFunc.C
## ~~~
##
## Original tutorial by Andras Zsenei.
## \macro_image
## \macro_code
##
## \author Alberto Ferro

import ROOT


f1a = ROOT.TF2("f1a","ROOT::Math::cauchy_pdf(x, y)",0,10,0,10)
f2a = ROOT.TF2("f2a","ROOT::Math::chisquared_pdf(x,y)",0,20, 0,20)
f3a = ROOT.TF2("f3a","ROOT::Math::gaussian_pdf(x,y)",0,10,0,5)
f4a = ROOT.TF2("f4a","ROOT::Math::tdistribution_pdf(x,y)",0,10,0,5)

c1 = ROOT.TCanvas("c1","c1",800,650)
c1.Divide(2,2)
c1.cd(1)
f1a.SetLineWidth(1)
f1a.Draw("surf1")
c1.cd(2)
f2a.SetLineWidth(1)
f2a.Draw("surf1")
c1.cd(3)
f3a.SetLineWidth(1)
f3a.Draw("surf1")
c1.cd(4)
f4a.SetLineWidth(1)
f4a.Draw("surf1")



# Author: Sergey Linev GSI  01/2025

################################################################################
# Copyright (C) 1995-2025, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

r"""
\pythondoc TCanvas

The TCanvas class is used to create the canvas on which graphical objects such as histograms can be drawn. The following
is a simple example of typical usage:

\code{.py}
import ROOT
import numpy


def plot():
    h = ROOT.TH1D("h", "h", 100, -5, 5)
    h.Fill(numpy.random.normal(size=1000))

    c = ROOT.TCanvas()
    h.Draw()
    c.Draw(block=True)


if __name__ == "__main__":
    print("Before plot function")
    plot()
    print("After plot function")
\endcode

Note the optional argument `block` passed to the `Draw` method of the canvas. If set to `True`, it will block the script
execution and run the ROOT graphics event loop until the <space> key is pressed. This allows interacting with the
canvas and its content until necessary, then move on with the rest of the script.

Another relevant use case is drawing live updates on a canvas, shown in the example below. In this case, the canvas is
setup by first creating the object to be drawn and drawing it once. Then, the object is updated in a for loop, which
could represent for example an incoming stream of data with which the histogram should be filled. Each time the plot
should be updated, the `ModifiedUpdate` function should be called. This will immediately show the new contents on the
plot. Finally, the canvas is drawn again with `Draw(block=True)` at the end of the loop so that it stays visible and can
be interacted with.

\code{.py}
import ROOT
import numpy


def live_update():
    c = ROOT.TCanvas()
    h = ROOT.TH1D("h", "h", 100, -5, 5)

    h.Draw()
    for _ in range(100):
        h.Fill(numpy.random.normal(size=10))
        c.ModifiedUpdate()
    c.Draw(block=True)

if __name__ == "__main__":
    print("Before plot function")
    live_update()
    print("After plot function")
\endcode

\endpythondoc
"""

from . import _run_root_event_loop, pythonization


def _TCanvas_init_releasing_ownership(self, *args, **kwargs):
    """
    Forward the arguments to the C++ constructor and give up Python ownership
    if the canvas registered itself to gROOT's list of canvases (which is
    the case for any non-default constructor). This avoids for example double deletes when a subsequent TCanvas with the same name
    triggers gROOT to delete the old canvas while the previous Python
    proxy is still alive somewhere.
    """
    import ROOT

    self._cpp_constructor(*args, **kwargs)
    if ROOT.gROOT.GetListOfCanvases().FindObject(self):
        ROOT.SetOwnership(self, False)


def _TCanvas_Update(self, block=False):
    """
    Updates the canvas.
    Also blocks script execution and runs the ROOT graphics event loop until the <space> key is pressed,
    but only if the following conditions are met:
    * The `block` optional argument is set to `True`.
    * ROOT graphics are enabled, i.e. `ROOT.gROOT.IsBatch() == False`.
    * The script is not running in ipython notebooks.
    """

    self._Update()

    # run loop if block flag is set
    if block:
        _run_root_event_loop()


def _TCanvas_Draw(self, option: str = "", block: bool = False):
    """
    Draw the canvas.
    Also blocks script execution and runs the ROOT graphics event loop until the <space> is pressed,
    but only if the following conditions are met:
    * The `block` optional argument is set to `True`.
    * ROOT graphics are enabled, i.e. `ROOT.gROOT.IsBatch() == False`.
    * The script is not running in ipython notebooks.
    """

    self._Draw(option)

    # run loop if block flag is set
    if block:
        self._Update()
        _run_root_event_loop()


@pythonization("TCanvas")
def pythonize_tcanvas(klass):
    # Parameters:
    # klass: class to be pythonized

    klass._Update = klass.Update
    klass._Draw = klass.Draw
    klass.Update = _TCanvas_Update
    klass.Draw = _TCanvas_Draw

    klass._cpp_constructor = klass.__init__
    klass.__init__ = _TCanvas_init_releasing_ownership



# Author: Stefan Wunsch CERN  08/2018

################################################################################
# Copyright (C) 1995-2018, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

r"""
\pythondoc ROOT::VecOps::RVec

The ROOT::RVec class has additional features in Python, which allow to adopt memory
from Numpy arrays and vice versa. The purpose of these features is the copyless
interfacing of Python and C++ using their most common data containers, Numpy arrays
and RVec with a std::vector interface.

### Conversion of RVecs to Numpy arrays

RVecs of fundamental types (int, float, ...) have in Python the `__array_interface__`
attribute attached. This information allows Numpy to adopt the memory of RVecs without
copying the content. You can find further documentation regarding the Numpy array interface
[here](https://numpy.org/doc/stable/reference/arrays.interface.html). The following code example
demonstrates the memory adoption mechanism using `numpy.asarray`.

\code{.py}
rvec = ROOT.RVec('double')((1, 2, 3))
print(rvec) # { 1.0000000, 2.0000000, 3.0000000 }

npy = numpy.asarray(rvec)
print(npy) # [1. 2. 3.]

rvec[0] = 42
print(npy) # [42. 2. 3.]
\endcode

### Conversion of Numpy arrays to RVecs

Data owned by Numpy arrays with fundamental types (int, float, ...) can be adopted by RVecs. To
create an RVec from a Numpy array, ROOT offers the facility ROOT.VecOps.AsRVec, which performs
a similar operation to `numpy.asarray`, but vice versa. A code example demonstrating the feature and
the adoption of the data owned by the Numpy array is shown below.

\code{.py}
npy = numpy.array([1.0, 2.0, 3.0])
print(npy) # [1. 2. 3.]

rvec = ROOT.VecOps.AsRVec(npy)
print(rvec) # { 1.0000000, 2.0000000, 3.0000000 }

npy[0] = 42
print(rvec) # { 42.000000, 2.0000000, 3.0000000 }
\endcode

\endpythondoc
"""

import sys

from . import pythonization

# This map includes all relevant C++ fundamental types found at
# https://en.cppreference.com/w/cpp/language/types.html and the associated
# ROOT portable types when available.
_array_interface_dtype_map = {
    # Integral types
    # C++ standard integer types
    "short": "i",
    "short int": "i",
    "signed short": "i",
    "signed short int": "i",
    "unsigned short": "u",
    "unsigned short int": "u",
    "int": "i",
    "signed": "i",
    "signed int": "i",
    "unsigned": "u",
    "unsigned int": "u",
    "long": "i",
    "long int": "i",
    "signed long": "i",
    "signed long int": "i",
    "unsigned long": "u",
    "unsigned long int": "u",
    "long long": "i",
    "long long int": "i",
    "signed long long": "i",
    "signed long long int": "i",
    "unsigned long long": "u",
    "unsigned long long int": "u",
    "std::size_t": "i",
    # Extended standard integer types
    "std::int8_t": "i",
    "std::int16_t": "i",
    "std::int32_t": "i",
    "std::int64_t": "i",
    "std::uint8_t": "u",
    "std::uint16_t": "u",
    "std::uint32_t": "u",
    "std::uint64_t": "u",
    # ROOT integer types
    "Int_t": "i",
    "UInt_t": "u",
    "Short_t": "i",
    "UShort_t": "u",
    "Long_t": "i",
    "ULong_t": "u",
    "Long64_t": "i",
    "ULong64_t": "u",
    # Boolean type
    "bool": "b",
    "Bool_t": "b",
    # Character types
    "char": "i",
    "Char_t": "i",
    "signed char": "i",
    "unsigned char": "u",
    "UChar_t": "u",
    "char16_t": "i",
    "char32_t": "i",
    # Floating-point types
    # C++ standard floating-point types
    "float": "f",
    "double": "f",
    "long double": "f",
    # ROOT floating-point types
    "Float_t": "f",
    "Double_t": "f",
}


def _get_cpp_type_from_numpy_type(dtype):
    cpptypes = {
        "i1": "signed char",
        "u1": "unsigned char",
        "i2": "Short_t",
        "u2": "UShort_t",
        "i4": "int",
        "u4": "unsigned int",
        "i8": "Long64_t",
        "u8": "ULong64_t",
        "f4": "float",
        "f8": "double",
        "b1": "bool",
    }

    if dtype not in cpptypes:
        raise RuntimeError("Object not convertible: Python object has unknown data-type '" + dtype + "'.")

    return cpptypes[dtype]


def _AsRVec(arr):
    r"""
    Adopt memory of a Python object with array interface using an RVec.

    \param[in] self self object
    \param[in] obj PyObject with array interface

    This function returns an RVec which adopts the memory of the given
    PyObject. The RVec takes the data pointer and the size from the array
    interface dictionary.
    Note that for arrays of strings and int8, the input data is copied into the RVec
    rather than adopted.
    """
    import math

    import numpy as np

    import ROOT

    # Get array interface of object
    interface = arr.__array_interface__

    # Get the data-pointer
    data = interface["data"][0]

    # Get the size of the contiguous memory
    shape = interface["shape"]
    size = math.prod(shape) if len(shape) > 0 else 0

    # Get the typestring and properties thereof
    typestr = interface["typestr"]
    dtype = typestr[1:]

    # Construct an RVec of strings by copying the input array
    if dtype == "O" or dtype.startswith("U"):
        underlying_object_types = {type(elem) for elem in arr}
        if len(underlying_object_types) > 1:
            raise TypeError(
                "All elements in the numpy array must be of the same type. Found types: {}".format(
                    underlying_object_types
                )
            )

        if underlying_object_types and underlying_object_types.pop() in [str, np.str_]:
            return ROOT.VecOps.RVec["std::string"](arr)
        else:
            raise TypeError("Cannot create an RVec from a numpy array of data type object.")

    # Construct an RVec of int8 by copying the input array
    # RVec<signed char> does not expose the ptr+size contructor
    if dtype == "i1":
        return ROOT.VecOps.RVec["signed char"](arr)

    if len(typestr) != 3:
        raise RuntimeError(
            "Object not convertible: __array_interface__['typestr'] returned '"
            + typestr
            + "' with invalid length unequal 3."
        )

    # Construct an RVec of the correct data-type
    cppdtype = _get_cpp_type_from_numpy_type(dtype)
    out = ROOT.VecOps.RVec[cppdtype](ROOT._cppyy.ll.reinterpret_cast[f"{cppdtype} *"](data), size)

    # Bind pyobject holding adopted memory to the RVec
    out.__adopted__ = arr

    return out


def get_array_interface(self):
    import ROOT

    cppname = type(self).__cpp_name__
    for dtype in _array_interface_dtype_map:
        if cppname.endswith("<{}>".format(dtype)):
            dtype_numpy = _array_interface_dtype_map[dtype]
            dtype_size = ROOT._cppyy.sizeof(dtype)
            endianness = "<" if sys.byteorder == "little" else ">"
            size = self.size()
            # Numpy breaks for data pointer of 0 even though the array is empty.
            # We set the pointer to 1 but the value itself is arbitrary and never accessed.
            if self.empty():
                pointer = 1
            else:
                pointer = ROOT._cppyy.ll.addressof(self.data())
            return {
                "shape": (size,),
                "typestr": "{}{}{}".format(endianness, dtype_numpy, dtype_size),
                "version": 3,
                "data": (pointer, False),
            }


def add_array_interface_property(klass, name):
    if True in [name.endswith("<{}>".format(dtype)) for dtype in _array_interface_dtype_map]:
        klass.__array_interface__ = property(get_array_interface)


@pythonization("RVec<", ns="ROOT::VecOps", is_prefix=True)
def pythonize_rvec(klass, name):
    # Parameters:
    # klass: class to be pythonized
    # name: string containing the name of the class

    # Add numpy array interface
    add_array_interface_property(klass, name)



## \file
## \ingroup tutorial_dataframe
## \notebook -draw
## Write ROOT data with RDataFrame.
##
## This tutorial shows how to write out datasets in ROOT format using RDataFrame.
##
## \macro_image
## \macro_code
##
## \date April 2017
## \author Danilo Piparo (CERN)

import ROOT

# A simple helper function to fill a test tree: this makes the example stand-alone.
def fill_tree(treeName, fileName):
    df = ROOT.RDataFrame(10000)
    df.Define("b1", "(int) rdfentry_")\
      .Define("b2", "(float) rdfentry_ * rdfentry_").Snapshot(treeName, fileName)

# We prepare an input tree to run on
fileName = "df007_snapshot_py.root"
outFileName = "df007_snapshot_output_py.root"
outFileNameAllColumns = "df007_snapshot_output_allColumns_py.root"
treeName = "myTree"
fill_tree(treeName, fileName)

# We read the tree from the file and create a RDataFrame
d = ROOT.RDataFrame(treeName, fileName)

# ## Select entries
# We now select some entries in the dataset
d_cut = d.Filter("b1 % 2 == 0")
# ## Enrich the dataset
# Build some temporary columns: we'll write them out

getVector_code ='''
std::vector<float> getVector (float b2)
{
   std::vector<float> v;
   for (int i = 0; i < 3; i++) v.push_back(b2*i);
   return v;
}
'''
ROOT.gInterpreter.Declare(getVector_code)

d2 = d_cut.Define("b1_square", "b1 * b1") \
          .Define("b2_vector", "getVector( b2 )")

# ## Write it to disk in ROOT format
# We now write to disk a new dataset with one of the variables originally
# present in the tree and the new variables.
# The user can explicitly specify the types of the columns as template
# arguments of the Snapshot method, otherwise they will be automatically
# inferred.
d2.Snapshot(treeName, outFileName, \
    ["b1", "b1_square", "b2_vector"])
# Open the new file and list the columns of the tree
f1 = ROOT.TFile(outFileName)
t = f1[treeName]
print("These are the columns b1, b1_square and b2_vector:")
for branch in t.GetListOfBranches():
    print("Branch: %s" %branch.GetName())

f1.Close()

# We are not forced to write the full set of column names. We can also
# specify a regular expression for that. In case nothing is specified, all
# columns are persistified.
d2.Snapshot(treeName, outFileNameAllColumns)

# Open the new file and list the columns of the tree
f2 = ROOT.TFile(outFileNameAllColumns)
t = f2[treeName]
print("These are all the columns available to this dataframe:")
for branch in t.GetListOfBranches():
    print("Branch: %s" %branch.GetName())

f2.Close()

# We can also get a fresh RDataFrame out of the snapshot and restart the
# analysis chain from it.

snapshot_df = d2.Snapshot(treeName, outFileName, ["b1_square"]);
h = snapshot_df.Histo1D("b1_square")

c = ROOT.TCanvas()
h.Draw()
c.SaveAs("df007_snapshot.png")

print("Saved figure to df007_snapshot.png")



## \file
## \ingroup tutorial_roofit_main
## \notebook -nodraw
## Addition and convolution: tools for visualization of ROOT.RooAbsArg expression trees
##
## \macro_code
## \macro_output
##
## \date February 2018
## \authors Clemens Lange, Wouter Verkerke (C++ version)

import ROOT

# Set up composite pdf
# --------------------------------------

# Declare observable x
x = ROOT.RooRealVar("x", "x", 0, 10)

# Create two Gaussian PDFs g1(x,mean1,sigma) anf g2(x,mean2,sigma) and
# their parameters
mean = ROOT.RooRealVar("mean", "mean of gaussians", 5)
sigma1 = ROOT.RooRealVar("sigma1", "width of gaussians", 0.5)
sigma2 = ROOT.RooRealVar("sigma2", "width of gaussians", 1)
sig1 = ROOT.RooGaussian("sig1", "Signal component 1", x, mean, sigma1)
sig2 = ROOT.RooGaussian("sig2", "Signal component 2", x, mean, sigma2)

# Sum the signal components into a composite signal pdf
sig1frac = ROOT.RooRealVar("sig1frac", "fraction of component 1 in signal", 0.8, 0.0, 1.0)
sig = ROOT.RooAddPdf("sig", "Signal", [sig1, sig2], [sig1frac])

# Build Chebychev polynomial pdf
a0 = ROOT.RooRealVar("a0", "a0", 0.5, 0.0, 1.0)
a1 = ROOT.RooRealVar("a1", "a1", -0.2, 0.0, 1.0)
bkg1 = ROOT.RooChebychev("bkg1", "Background 1", x, [a0, a1])

# Build expontential pdf
alpha = ROOT.RooRealVar("alpha", "alpha", -1)
bkg2 = ROOT.RooExponential("bkg2", "Background 2", x, alpha)

# Sum the background components into a composite background pdf
bkg1frac = ROOT.RooRealVar("bkg1frac", "fraction of component 1 in background", 0.2, 0.0, 1.0)
bkg = ROOT.RooAddPdf("bkg", "Signal", [bkg1, bkg2], [bkg1frac])

# Sum the composite signal and background
bkgfrac = ROOT.RooRealVar("bkgfrac", "fraction of background", 0.5, 0.0, 1.0)
model = ROOT.RooAddPdf("model", "g1+g2+a", [bkg, sig], [bkgfrac])

# Print composite tree in ASCII
# -----------------------------------------------------------

# Print tree to stdout
model.Print("t")

# Print tree to file
model.printCompactTree("", "rf206_asciitree.txt")

# Draw composite tree graphically
# -------------------------------------------------------------

# Print GraphViz DOT file with representation of tree
model.graphVizTree("rf206_model.dot")

# Make graphic output file with one of the GraphViz tools
# (freely available from www.graphviz.org)
#
# 'Top-to-bottom graph'
# unix> dot -Tgif -o rf207_model_dot.gif rf207_model.dot
#
# 'Spring-model graph'
# unix> fdp -Tgif -o rf207_model_fdp.gif rf207_model.dot



# \file
# \ingroup tutorial_roostats
# \notebook
# Standard tutorial macro for performing an inverted  hypothesis test for computing an interval
#
# This macro will perform a scan of the p-values for computing the interval or limit
#
# Usage:
#
# ~~~{.py}
# ipython3> %run StandardHypoTestInvDemo.C
# ipython3> StandardHypoTestInvDemo("fileName","workspace name","S+B modelconfig name","B model name","data set")
# name",calculator type, test statistic type, use CLS,
#                                number of points, xmin, xmax, number of toys, use number counting)
#
# type = 0 Freq calculator
# type = 1 Hybrid calculator
# type = 2 Asymptotic calculator
# type = 3 Asymptotic calculator using nominal Asimov data sets (not using fitted parameter values but nominal ones)
#
# testStatType = 0 LEP
#              = 1 Tevatron
#              = 2 Profile Likelihood two sided
#              = 3 Profile Likelihood one sided (i.e. = 0 if mu < mu_hat)
#              = 4 Profile Likelihood signed ( pll = -pll if mu < mu_hat)
#              = 5 Max Likelihood Estimate as test statistic
#              = 6 Number of observed event as test statistic
# ~~~
#
# \macro_image
# \macro_output
# \macro_code
#
# \author Lorenzo Moneta (C++ version), and P. P. (Python translation)

import ROOT
import os


# structure defining the options
class HypoTestInvOptions:

    plotHypoTestResult = True  # plot test statistic result at each point
    writeResult = True  # write HypoTestInverterResult in a file
    resultFileName = (
        ROOT.TString()
    )  # file with results (by default is built automatically using the workspace input file name)
    optimize = True  # optimize evaluation of test statistic
    useVectorStore = True  # convert data to use  roofit data store
    generateBinned = False  # generate binned data sets
    noSystematics = False  # force all systematics to be off (i.e. set all nuisance parameters as constat
    # to their nominal values)
    nToysRatio = 2  # ratio Ntoys S+b/ntoysB
    maxPOI = -1  # max value used of POI (in case of auto scan)
    enableDetailedOutput = (
        False  # enable detailed output with all fit information for each toys (output will be written in result file)
    )
    rebuild = False  # re-do extra toys for computing expected limits and rebuild test stat
    # distributions (N.B this requires much more CPU (factor is equivalent to nToyToRebuild)
    nToyToRebuild = 100  # number of toys used to rebuild
    rebuildParamValues = 0  # = 0 do a profile of all the parameters on the B (alt snapshot) before performing a
    # rebuild operation (default)
    # = 1   use initial workspace parameters with B snapshot values
    # = 2   use all initial workspace parameters with B
    # Otherwise the rebuild will be performed using
    initialFit = -1  # do a first fit to the model (-1 : default, 0 skip fit, 1 do always fit)
    randomSeed = -1  # random seed (if = -1: use default value, if = 0 always random )

    nAsimovBins = 0  # number of bins in observables used for Asimov data sets (0 is the default and it is given by
    # workspace, typically is 100)

    reuseAltToys = False  # reuse same toys for alternate hypothesis (if set one gets more stable bands)
    confLevel = 0.95  # confidence level value

    minimizerType = ""  # minimizer type (default is what is in ROOT.Math.MinimizerOptions.DefaultMinimizerType()
    massValue = ""  # extra string to tag output file of result
    printLevel = 0  # print level for debugging PL test statistics and calculators

    useNLLOffset = False  # use NLL offset when fitting (this increase stability of fits)


optHTInv = HypoTestInvOptions()

# internal class to run the inverter and more

large_declaration = """ \
namespace RooStats{

class HypoTestInvTool {

   public:
   HypoTestInvTool();
   ~HypoTestInvTool(){};

   HypoTestInverterResult *RunInverter(RooWorkspace *w, const char *modelSBName, const char *modelBName,
   const char *dataName, int type, int testStatType, bool useCLs, int npoints,
   double poimin, double poimax, int ntoys, bool useNumberCounting = false,
   const char *nuisPriorName = 0);

   void AnalyzeResult(HypoTestInverterResult *r, int calculatorType, int testStatType, bool useCLs, int npoints,
   const char *fileNameBase = 0);

   void SetParameter(const char *name, const char *value);
   void SetParameter(const char *name, bool value);
   void SetParameter(const char *name, int value);
   void SetParameter(const char *name, double value);

   //private:
   public:
   bool mPlotHypoTestResult;
   bool mWriteResult;
   bool mOptimize;
   bool mUseVectorStore;
   bool mGenerateBinned;
   bool mRebuild;
   bool mReuseAltToys;
   bool mEnableDetOutput;
   int mNToyToRebuild;
   int mRebuildParamValues;
   int mPrintLevel;
   int mInitialFit;
   int mRandomSeed;
   double mNToysRatio;
   double mMaxPoi;
   int mAsimovBins;
   std::string mMassValue;
   std::string
   mMinimizerType; // minimizer type (default is what is in ROOT::Math::MinimizerOptions::DefaultMinimizerType())
   TString mResultFileName;
};

} // end using RooStats namespace

RooStats::HypoTestInvTool::HypoTestInvTool()
   : mPlotHypoTestResult(true), mWriteResult(false), mOptimize(true), mUseVectorStore(true), mGenerateBinned(false),
     mEnableDetOutput(false), mRebuild(false), mReuseAltToys(false),
     mNToyToRebuild(100), mRebuildParamValues(0), mPrintLevel(0), mInitialFit(-1), mRandomSeed(-1), mNToysRatio(2),
     mMaxPoi(-1), mAsimovBins(0), mMassValue(""), mMinimizerType(""), mResultFileName()
{
}

"""
ROOT.gInterpreter.Declare(large_declaration)


# Expanding definitions of class: HypoTestInvTool into HypoTestInvTool_plus
class HypoTestInvTool_plus(ROOT.RooStats.HypoTestInvTool):
    # RooStats.HypoTestInvTool.SetParameter(name,  value):
    def SetParameter(self, name, value):  # SetParameter is polymorphic.
        if (type(name) is str) and (type(value) is str):
            #
            # set boolean parameters
            #

            s_name = ROOT.std.string(name)

            if s_name.find("PlotHypoTestResult") != ROOT.std.string.npos:
                self.mPlotHypoTestResult = value
            if s_name.find("WriteResult") != ROOT.std.string.npos:
                self.mWriteResult = value
            if s_name.find("Optimize") != ROOT.std.string.npos:
                self.mOptimize = value
            if s_name.find("UseVectorStore") != ROOT.std.string.npos:
                self.mUseVectorStore = value
            if s_name.find("GenerateBinned") != ROOT.std.string.npos:
                self.mGenerateBinned = value
            if s_name.find("EnableDetailedOutput") != ROOT.std.string.npos:
                self.mEnableDetOutput = value
            if s_name.find("Rebuild") != ROOT.std.string.npos:
                self.mRebuild = value
            if s_name.find("ReuseAltToys") != ROOT.std.string.npos:
                self.mReuseAltToys = value

            return
        # RooStats.HypoTestInvTool.SetParameter = SetParameter

        # RooStats.HypoTestInvTool.SetParameter(name, value):
        # def SetParameter(self, name, value):
        elif (type(name) is str) and (type(value) is bool):

            #
            # set integer parameters
            #

            s_name = ROOT.std.string(name)

            if s_name.find("NToyToRebuild") != ROOT.std.string.npos:
                self.mNToyToRebuild = value
            if s_name.find("RebuildParamValues") != ROOT.std.string.npos:
                self.mRebuildParamValues = value
            if s_name.find("PrintLevel") != ROOT.std.string.npos:
                self.mPrintLevel = value
            if s_name.find("InitialFit") != ROOT.std.string.npos:
                self.mInitialFit = value
            if s_name.find("RandomSeed") != ROOT.std.string.npos:
                self.mRandomSeed = value
            if s_name.find("AsimovBins") != ROOT.std.string.npos:
                self.mAsimovBins = value

            return
        # RooStats.HypoTestInvTool.SetParameter = SetParameter

        # RooStats.HypoTestInvTool.SetParameter(name, value):
        # def SetParameter(self, name, value):
        elif (type(name) is str) and (type(value) is int):

            #
            # set double precision parameters
            #

            s_name = ROOT.std.string(name)

            if s_name.find("NToysRatio") != ROOT.std.string.npos:
                self.mNToysRatio = value
            if s_name.find("MaxPOI") != ROOT.std.string.npos:
                self.mMaxPoi = value

            return
        # RooStats.HypoTestInvTool.SetParameter = SetParameter

        # RooStats.HypoTestInvTool.SetParameter(name, value):
        # def SetParameter(self, name, value):
        elif (type(name) is str) and (type(value) is (float or double)):

            #
            # set string parameters
            #

            s_name = ROOT.std.string(name)

            if s_name.find("MassValue") != ROOT.std.string.npos:
                global gselfmMassValue
                gselfmMassValue = self.mMassValue
                self.mMassValue = value  # (self.mMassValue).assign(value)
            if s_name.find("MinimizerType") != ROOT.std.string.npos:
                self.mMassValue = value  # self.mMinimizerType.assign(value)
            if s_name.find("ResultFileName") != ROOT.std.string.npos:
                self.mResultFileName = value

            return

    ROOT.RooStats.HypoTestInvTool.SetParameter = SetParameter

    # --piece of code moved forwards...-
    # ---

    # RooStats.HypoTestInvTool.AnalyzeResult
    def AnalyzeResult(self, r, calculatorType, testStatType, useCLs, npoints, fileNameBase):
        # type(r) is HypoTestInverterResult

        # analyze result produced by the inverter, optionally save it in a file

        lowerLimit = 0
        llError = 0
        # if defined ROOT_SVN_VERSION && ROOT_SVN_VERSION >= 44126
        if r.IsTwoSided():
            lowerLimit = r.LowerLimit()
            llError = r.LowerLimitEstimatedError()

        # else
        lowerLimit = r.LowerLimit()
        llError = r.LowerLimitEstimatedError()
        # endif

        upperLimit = r.UpperLimit()
        ulError = r.UpperLimitEstimatedError()

        # ROOT.std::cout << "DEBUG : [ " << lowerLimit << " , " << upperLimit << "  ] " << ROOT.std::endl;

        if lowerLimit < upperLimit * (1.0 - 1.0e-4) and lowerLimit != 0:
            print(f"The computed lower limit is: ", lowerLimit, " +/- ", llError)

        print(f"The computed upper limit is: {upperLimit} +/- ", ulError)

        # compute expected limit
        print(f"Expected upper limits, using the B (alternate) model : ")
        print(f" expected limit (median) ", r.GetExpectedUpperLimit(0))
        print(f" expected limit (-1 sig) ", r.GetExpectedUpperLimit(-1))
        print(f" expected limit (+1 sig) ", r.GetExpectedUpperLimit(1))
        print(f" expected limit (-2 sig) ", r.GetExpectedUpperLimit(-2))
        print(f" expected limit (+2 sig) ", r.GetExpectedUpperLimit(2))

        # detailed output
        if self.mEnableDetOutput:
            self.mWriteResult = True
            ROOT.Info("StandardHypoTestInvDemo", "detailed output will be written in output result file")

        # write result in a file
        if r != ROOT.kNone and self.mWriteResult:

            # write to a file the results
            calcType = "Freq" if (calculatorType == 0) else ("Hybr" if (calculatorType == 1) else "Asym")
            limitType = "CLs" if (useCLs) else "Cls+b"
            scanType = "auto" if (npoints < 0) else "grid"
            if self.mResultFileName.IsNull():
                self.mResultFileName = ROOT.TString.Format(
                    "%s_%s_%s_ts%d_", calcType, limitType, scanType, testStatType
                )
                # strip the / from the filename
                if self.mMassValue.size() > 0:
                    self.mResultFileName += self.mMassValue
                    self.mResultFileName += "_"

                name = fileNameBase
                name.Replace(0, name.Last("/") + 1, "")
                self.mResultFileName += name

            # get (if existing) rebuilt UL distribution
            uldistFile = "RULDist.root"
            ulDist = 0
            existULDist = not ROOT.gSystem.AccessPathName(uldistFile)
            if existULDist:
                fileULDist = TFile.Open(uldistFile)
                if fileULDist:
                    ulDist = fileULDist.Get("RULDist")

            fileOut = TFile(self.mResultFileName, "RECREATE")
            r.Write()
            if ulDist:
                ulDist.Write()
            ROOT.Info(
                "StandardHypoTestInvDemo",
                "HypoTestInverterResult has been written in the file %s",
                self.mResultFileName.Data(),
            )

            fileOut.Close()

        # plot the result ( p values vs scan points)
        typeName = ""
        if calculatorType == 0:
            typeName = "Frequentist"
        if calculatorType == 1:
            typeName = "Hybrid"
        elif calculatorType == 2 or calculatorType == 3:
            typeName = "Asymptotic"
            self.mPlotHypoTestResult = False

        resultName = r.GetName()
        plotTitle = "{} CL Scan for workspace {}".format(str(typeName), resultName)
        global gr
        gr = r
        plot = ROOT.RooStats.HypoTestInverterPlot("HTI_Result_Plot", plotTitle, r)

        # plot in a new canvas with style
        c1Name = "{}_Scan".format(typeName)
        c1 = ROOT.TCanvas(c1Name)
        c1.SetLogy(False)

        plot.Draw("CLb 2CL")  # plot all and Clb
        # if (useCLs):
        #    plot.Draw("CLb 2CL")  # plot all and Clb ???
        # else:
        #    plot.Draw("")  # plot all and Clb ???
        c1.Update()
        c1.Draw()
        c1.SaveAs("StandardHypoTestInvDemo.c1.png")

        nEntries = r.ArraySize()

        # plot test statistics distributions for the two hypothesis
        if self.mPlotHypoTestResult:
            c2 = ROOT.TCanvas("c2")
            if nEntries > 1:
                ny = ROOT.TMath.CeilNint(ROOT.TMath.Sqrt(nEntries))
                nx = ROOT.TMath.CeilNint(float(nEntries) / ny)
                c2.Divide(nx, ny)

            for i in range(nEntries):
                if nEntries > 1:
                    c2.cd(i + 1)
                pl = plot.MakeTestStatPlot(i)
                pl.SetLogYaxis(True)
                pl.Draw()
            c2.Update()
            c2.Draw()
            c2.SaveAs("StandardHypoTestInvDemo.c2.png")

        ROOT.gPad = c1

    ROOT.RooStats.HypoTestInvTool.AnalyzeResult = AnalyzeResult

    # internal routine to run the inverter
    # RooStats.HypoTestInvTool.RunInverter
    def RunInverter(
        self,
        w,
        modelSBName,
        modelBName,
        dataName,
        Type,
        testStatType,
        useCLs,
        npoints,
        poimin,
        poimax,
        ntoys,
        useNumberCounting,
        nuisPriorName,
    ):

        print(f"Running HypoTestInverter on the workspace ", w.GetName())

        w.Print()

        data = w.data(dataName)
        if not data:
            Error("StandardHypoTestDemo", "Not existing data {}".format(dataName))
            return 0
        else:
            print(f"Using data set ", dataName)

        if self.mUseVectorStore:
            ROOT.RooAbsData.setDefaultStorageType(ROOT.RooAbsData.Vector)
            data.convertToVectorStore()

        # get models from WS
        # get the modelConfig out of the file
        bModel = w.obj(modelBName)
        sbModel = w.obj(modelSBName)

        if not sbModel:
            Error("StandardHypoTestDemo", "Not existing ModelConfig %s", modelSBName)
            return 0

        # check the model
        if not sbModel.GetPdf():
            Error("StandardHypoTestDemo", "Model %s has no pdf ", modelSBName)
            return 0

        if not sbModel.GetParametersOfInterest():
            Error("StandardHypoTestDemo", "Model %s has no poi ", modelSBName)
            return 0

        if not sbModel.GetObservables():
            Error("StandardHypoTestInvDemo", "Model %s has no observables ", modelSBName)
            return 0

        if not sbModel.GetSnapshot():
            ROOT.Info(
                "StandardHypoTestInvDemo", "Model {} has no snapshot  - make one using model poi".format(modelSBName)
            )
            sbModel.SetSnapshot(sbModel.GetParametersOfInterest())

        # case of no systematics
        # remove nuisance parameters from model
        if optHTInv.noSystematics:
            nuisPar = sbModel.GetNuisanceParameters()
            if nuisPar and nuisPar.getSize() > 0:
                print(
                    f"StandardHypoTestInvDemo",
                    "  -  Switch off all systematics by setting them constant to their initial values",
                )
                ROOT.RooStats.SetAllConstant(nuisPar)

            if bModel:
                bnuisPar = bModel.GetNuisanceParameters()
                if bnuisPar:
                    ROOT.RooStats.SetAllConstant(bnuisPar)

        if (not bModel) or (bModel == sbModel):
            ROOT.Info("StandardHypoTestInvDemo", "The background model  {} does not exist".format(modelBName))
            ROOT.Info("StandardHypoTestInvDemo", "Copy it from ModelConfig {} and set POI to zero".format(modelSBName))
            bModel = sbModel.Clone()
            bModel.SetName(modelSBName + "_with_poi_0")
            var = bModel.GetParametersOfInterest().first()
            if not var:
                return 0
            oldval = var.getVal()
            var.setVal(0)
            bModel.SetSnapshot(ROOT.RooArgSet(var))
            var.setVal(oldval)
        else:
            if not bModel.GetSnapshot():
                ROOT.Info(
                    "StandardHypoTestInvDemo",
                    "Model %s has no snapshot  - make one using model poi and 0 values ",
                    modelBName,
                )
                var = bModel.GetParametersOfInterest().first()
                if var:
                    oldval = var.getVal()
                    var.setVal(0)
                    bModel.SetSnapshot(ROOT.RooArgSet(var))
                    var.setVal(oldval)
                else:
                    Error("StandardHypoTestInvDemo", "Model %s has no valid poi", modelBName)
                    return 0

        # check model  has global observables when there are nuisance pdf
        # for the hybrid case the globals are not needed
        if Type != 1:
            hasNuisParam = sbModel.GetNuisanceParameters() and sbModel.GetNuisanceParameters().getSize() > 0
            hasGlobalObs = sbModel.GetGlobalObservables() and sbModel.GetGlobalObservables().getSize() > 0
            if hasNuisParam and not hasGlobalObs:
                # try to see if model has nuisance parameters first
                constrPdf = RooStats.MakeNuisancePdf(sbModel, "nuisanceConstraintPdf_sbmodel")
                if constrPdf:
                    Warning(
                        "StandardHypoTestInvDemo",
                        "Model %s has nuisance parameters but no global observables associated",
                        sbModel.GetName(),
                    )
                    Warning(
                        "StandardHypoTestInvDemo",
                        "\tThe effect of the nuisance parameters will not be treated correctly ",
                    )

        # save all initial parameters of the model including the global observables
        initialParameters = ROOT.RooArgSet()
        allParams = sbModel.GetPdf().getParameters(data)
        allParams.snapshot(initialParameters)

        # run first a data fit

        poiSet = sbModel.GetParametersOfInterest()
        poi = poiSet.first()

        print("StandardHypoTestInvDemo : POI initial value:   ", poi.GetName(), " = ", poi.getVal())

        # fit the data first (need to use constraint )
        tw = ROOT.TStopwatch()

        doFit = self.mInitialFit
        if testStatType == 0 and self.mInitialFit == -1:
            doFit = False  # case of LEP test statistic
        if type == 3 and self.mInitialFit == -1:
            doFit = False  # case of Asymptoticcalculator with nominal Asimov
        poihat = 0

        if len(self.mMinimizerType) == 0:
            self.mMinimizerType = ROOT.Math.MinimizerOptions.DefaultMinimizerType()
        else:
            ROOT.Math.MinimizerOptions.SetDefaultMinimizer(str(self.mMinimizerType))

        ROOT.Info(
            "StandardHypoTestInvDemo",
            "Using {} as minimizer for computing the test statistic".format(
                str(ROOT.Math.MinimizerOptions.DefaultMinimizerType())
            ),
        )

        if doFit:

            # do the fit : By doing a fit the POI snapshot (for S+B)  is set to the fit value
            # and the nuisance parameters nominal values will be set to the fit value.
            # This is relevant when using LEP test statistics

            ROOT.Info("StandardHypoTestInvDemo", " Doing a first fit to the observed data ")
            constrainParams = RooArgSet()
            if sbModel.GetNuisanceParameters():
                constrainParams.add(sbModel.GetNuisanceParameters())
            RooStats.RemoveConstantParameters(constrainParams)
            tw.Start()
            fitres = sbModel.GetPdf().fitTo(
                data,
                InitialHesse(False),
                Hesse(False),
                Minimizer(str(self.mMinimizerType), "Migrad"),
                Strategy(0),
                PrintLevel(self.mPrintLevel),
                Constrain(constrainParams),
                Save(True),
                Offset(RooStats.NLLOffsetMode()),
            )

            if fitres.status() != 0:
                Warning(
                    "StandardHypoTestInvDemo",
                    "Fit to the model failed - try with strategy 1 and perform first an Hesse computation",
                )
                fitres = sbModel.GetPdf().fitTo(
                    data,
                    InitialHesse(True),
                    Hesse(False),
                    Minimizer(self.mMinimizerType, "Migrad"),
                    Strategy(1),
                    PrintLevel(self.mPrintLevel + 1),
                    Constrain(constrainParams),
                    Save(True),
                    Offset(RooStats.NLLOffsetMode()),
                )

            if fitres.status() != 0:
                Warning("StandardHypoTestInvDemo", " Fit still failed - continue anyway.....")

            print("StandardHypoTestInvDemo - Best Fit value : ", poi.GetName(), " = ", poihat, " +/- ", poi.getError())
            print(f"Time for fitting : ")
            tw.Print()

            # save best fit value in the poi snapshot
            sbModel.SetSnapshot(sbModel.GetParametersOfInterest())
            print(f"StandardHypoTestInvo: snapshot of S+B Model ", sbModel.GetName(), " is set to the best fit value")

        # print a message in case of LEP test statistics because it affects result by doing or not doing a fit
        if testStatType == 0:
            if not doFit:
                ROOT.Info(
                    "StandardHypoTestInvDemo",
                    "Using LEP test statistic - an initial fit is not done and the TS will use "
                    + "the nuisances at the model value",
                )
            else:
                ROOT.Info(
                    "StandardHypoTestInvDemo",
                    "Using LEP test statistic - an initial fit has been done and the TS will use "
                    + "the nuisances at the best fit value",
                )

        # build test statistics and hypotest calculators for running the inverter

        slrts = ROOT.RooStats.SimpleLikelihoodRatioTestStat(sbModel.GetPdf(), bModel.GetPdf())

        # null parameters must includes snapshot of poi plus the nuisance values
        nullParams = ROOT.RooArgSet(sbModel.GetSnapshot())
        if sbModel.GetNuisanceParameters():
            nullParams.add(sbModel.GetNuisanceParameters())
        if sbModel.GetSnapshot():
            slrts.SetNullParameters(nullParams)
            altParams = ROOT.RooArgSet(bModel.GetSnapshot())
        if bModel.GetNuisanceParameters():
            altParams.add(bModel.GetNuisanceParameters())
        if bModel.GetSnapshot():
            slrts.SetAltParameters(altParams)
        if self.mEnableDetOutput:
            slrts.EnableDetailedOutput()

        # ratio of profile likelihood - need to pass snapshot for the alt
        ropl = ROOT.RooStats.RatioOfProfiledLikelihoodsTestStat(sbModel.GetPdf(), bModel.GetPdf(), bModel.GetSnapshot())
        ropl.SetSubtractMLE(False)
        if testStatType == 11:
            ropl.SetSubtractMLE(True)
        ropl.SetPrintLevel(self.mPrintLevel)
        ropl.SetMinimizer(self.mMinimizerType.c_str())
        if self.mEnableDetOutput:
            ropl.EnableDetailedOutput()

        profll = ROOT.RooStats.ProfileLikelihoodTestStat(sbModel.GetPdf())
        if testStatType == 3:
            profll.SetOneSided(True)
        if testStatType == 4:
            profll.SetSigned(True)
        profll.SetMinimizer(self.mMinimizerType.c_str())
        profll.SetPrintLevel(self.mPrintLevel)
        if self.mEnableDetOutput:
            profll.EnableDetailedOutput()

        profll.SetReuseNLL(self.mOptimize)
        slrts.SetReuseNLL(self.mOptimize)
        ropl.SetReuseNLL(self.mOptimize)

        if self.mOptimize:
            profll.SetStrategy(0)
            ropl.SetStrategy(0)
            ROOT.Math.MinimizerOptions.SetDefaultStrategy(0)

        if self.mMaxPoi > 0:
            poi.setMax(self.mMaxPoi)  # increase limit

        maxll = ROOT.RooStats.MaxLikelihoodEstimateTestStat(sbModel.GetPdf(), poi)
        nevtts = ROOT.RooStats.NumEventsTestStat()

        ROOT.RooStats.AsymptoticCalculator.SetPrintLevel(self.mPrintLevel)

        # create the HypoTest calculator class
        hc = ROOT.nullptr
        if Type == 0:
            hc = ROOT.RooStats.FrequentistCalculator(data, bModel, sbModel)
        elif Type == 1:
            hc = ROOT.RooStats.HybridCalculator(data, bModel, sbModel)
        # elif (Type == 2 ):
        #    hc = AsymptoticCalculator(data, bModel, sbModel, false, self.mAsimovBins)
        # elif (Type == 3 ):
        #    hc = AsymptoticCalculator(data, bModel, sbModel, True, self.mAsimovBins)  # for using
        # Asimov data generated with nominal values
        elif Type == 2:
            hc = ROOT.RooStats.AsymptoticCalculator(data, bModel, sbModel, False)
        elif Type == 3:
            hc = ROOT.RooStats.AsymptoticCalculator(
                data, bModel, sbModel, True
            )  # for using Asimov data generated with nominal values
        else:
            Error(
                "StandardHypoTestInvDemo",
                "Invalid - calculator type = {Type}supported values are only :\n\t\t\t 0 "
                + "(Frequentist) , 1 (Hybrid) , 2 (Asymptotic) ",
            )

            return 0

        # set the test statistic
        testStat = 0
        if testStatType == 0:
            testStat = slrts
        if testStatType == 1 or testStatType == 11:
            testStat = ropl
        if testStatType == 2 or testStatType == 3 or testStatType == 4:
            testStat = profll
        if testStatType == 5:
            testStat = maxll
        if testStatType == 6:
            testStat = nevtts

        if testStat == 0:
            Error(
                "StandardHypoTestInvDemo",
                "Invalid - test statistic type = {testStatType} supported values are only :\n\t\t\t 0 (SLR) "
                + ", 1 (Tevatron) , 2 (PLR), 3 (PLR1), 4(MLE)",
            )

            return 0

        toymcs = hc.GetTestStatSampler()
        if toymcs and (Type == 0 or Type == 1):
            # look if pdf is number counting or extended
            if sbModel.GetPdf().canBeExtended():
                if useNumberCounting:
                    Warning("StandardHypoTestInvDemo", "Pdf is extended: but number counting flag is set: ignore it ")
            else:
                # for not extended pdf
                if not useNumberCounting:
                    nEvents = data.numEntries()
                    ROOT.Info(
                        "StandardHypoTestInvDemo",
                        "Pdf is not extended: number of events to generate taken  from observed data set is {nEvents}",
                    )
                    toymcs.SetNEventsPerToy(nEvents)
                else:
                    ROOT.Info("StandardHypoTestInvDemo", "using a number counting pdf")
                    toymcs.SetNEventsPerToy(1)

            toymcs.SetTestStatistic(testStat)

            if data.isWeighted() and not self.mGenerateBinned:
                ROOT.Info(
                    "StandardHypoTestInvDemo",
                    (
                        "Data set is weighted, nentries = {} and sum of weights = {:8.1f} but toy"
                        + "generation is unbinned - it would be faster to set self.mGenerateBinned to true\n"
                    ).format(data.numEntries(), data.sumEntries()),
                )

            toymcs.SetGenerateBinned(self.mGenerateBinned)

            toymcs.SetUseMultiGen(self.mOptimize)

            if self.mGenerateBinned and sbModel.GetObservables().getSize() > 2:
                Warning(
                    "StandardHypoTestInvDemo",
                    (
                        "generate binned is activated but the number of observable is {}. Too much "
                        + "memory could be needed for allocating all the bins"
                    ).format(sbModel.GetObservables().getSize()),
                )

            # set the random seed if needed
            if self.mRandomSeed >= 0:
                RooRandom.randomGenerator().SetSeed(self.mRandomSeed)

        # specify if need to re-use same toys
        if self.mReuseAltToys:
            hc.UseSameAltToys()

        if Type == 1:
            hhc = HybridCalculator(hc)
            assert hhc

            hhc.SetToys(ntoys, ntoys / self.mNToysRatio)  # can use less ntoys for b hypothesis

            # remove global observables from ModelConfig (this is probably not needed anymore in 5.32)
            bModel.SetGlobalObservables(RooArgSet())
            sbModel.SetGlobalObservables(RooArgSet())

            # check for nuisance prior pdf in case of nuisance parameters
            if bModel.GetNuisanceParameters() or sbModel.GetNuisanceParameters():

                # fix for using multigen (does not work in this case)
                toymcs.SetUseMultiGen(False)
                ToyMCSampler.SetAlwaysUseMultiGen(False)

                nuisPdf = 0
                if nuisPriorName:
                    nuisPdf = w.pdf(nuisPriorName)
                # use prior defined first in bModel (then in SbModel)
                if not nuisPdf:
                    ROOT.Info(
                        "StandardHypoTestInvDemo",
                        "No nuisance pdf given for the HybridCalculator - try to deduce  pdf from the model",
                    )
                    if bModel.GetPdf() and bModel.GetObservables():
                        nuisPdf = RooStats.MakeNuisancePdf(bModel, "nuisancePdf_bmodel")
                    else:
                        nuisPdf = RooStats.MakeNuisancePdf(sbModel, "nuisancePdf_sbmodel")

                if not nuisPdf:
                    if bModel.GetPriorPdf():
                        nuisPdf = bModel.GetPriorPdf()
                        ROOT.Info(
                            "StandardHypoTestInvDemo",
                            "No nuisance pdf given - try to use %s that is defined as a prior pdf in the B model",
                            nuisPdf.GetName(),
                        )
                    else:
                        Error(
                            "StandardHypoTestInvDemo",
                            "Cannot run Hybrid calculator because no prior on the nuisance "
                            "parameter is specified or can be derived",
                        )
                        return 0

                assert nuisPdf
                ROOT.Info("StandardHypoTestInvDemo", "Using as nuisance Pdf ... ")
                nuisPdf.Print()

                nuisParams = (
                    ibModel.GetNuisanceParameters()
                    if (bModel.GetNuisanceParameters())
                    else (sbModel.GetNuisanceParameters())
                )
                npnuisPdf.getObservables(nuisParams)
                if np.getSize() == 0:
                    Warning(
                        "StandardHypoTestInvDemo",
                        "Prior nuisance does not depend on nuisance parameters. They will be smeared in their full range",
                    )

                hhc.ForcePriorNuisanceAlt(nuisPdf)
                hhc.ForcePriorNuisanceNull(nuisPdf)

        elif type == 2 or type == 3:
            if testStatType == 3:
                hc.SetOneSided(True)
            if testStatType != 2 and testStatType != 3:
                Warning(
                    "StandardHypoTestInvDemo",
                    "Only the PL test statistic can be used with AsymptoticCalculator - use by default a two-sided PL",
                )
        elif type == 0:
            hc.SetToys(ntoys, ntoys / self.mNToysRatio)
            # store also the fit information for each poi point used by calculator based on toys
            if self.mEnableDetOutput:
                hc.StoreFitInfo(True)
        elif type == 1:
            hc.SetToys(ntoys, ntoys / self.mNToysRatio)
            # store also the fit information for each poi point used by calculator based on toys
            # if (self.mEnableDetOutput):
            #    hc.StoreFitInfo(True)

        # Get the result
        ROOT.RooMsgService.instance().getStream(1).removeTopic(ROOT.RooFit.NumericIntegration)

        calc = ROOT.RooStats.HypoTestInverter(hc)
        calc.SetConfidenceLevel(optHTInv.confLevel)

        calc.UseCLs(useCLs)
        calc.SetVerbose(True)

        if npoints > 0:
            if poimin > poimax:
                # if no min/max given scan between MLE and +4 sigma
                poimin = int(poihat)
                poimax = int(poihat + 4 * poi.getError())

            print(f"Doing a fixed scan  in interval : {poimin}, {poimax} ")
            calc.SetFixedScan(npoints, poimin, poimax)
        else:
            # poi.setMax(10*int( (poihat+ 10 *poi.getError() )/10 ) )
            print(f"Doing an  automatic scan  in interval : {poi.getMin()} , ", poi.getMax())

        tw.Start()
        r = calc.GetInterval()
        print(f"Time to perform limit scan \n")
        tw.Print()

        if self.mRebuild:

            print(f"\n\n")
            print(
                f"Rebuild the upper limit distribution by re-generating new set of pseudo-experiment and re-compute "
                + "for each of them a new upper limit\n\n"
            )

            allParams = sbModel.GetPdf().getParameters(data)

            # define on which value of nuisance parameters to do the rebuild
            # default is best fit value for bmodel snapshot

            if self.mRebuildParamValues != 0:
                # set all parameters to their initial workspace values
                allParams.assign(initialParameters)

            if self.mRebuildParamValues == 0 or self.mRebuildParamValues == 1:
                constrainParams = RooArgSet()
                if sbModel.GetNuisanceParameters():
                    constrainParams.add(sbModel.GetNuisanceParameters())
                RooStats.RemoveConstantParameters(constrainParams)

                poiModel = sbModel.GetParametersOfInterest()
                bModel.LoadSnapshot()

                # do a profile using the B model snapshot
                if self.mRebuildParamValues == 0:

                    RooStats.SetAllConstant(poiModel, True)

                    sbModel.GetPdf().fitTo(
                        data,
                        InitialHesse(False),
                        Hesse(False),
                        Minimizer(self.mMinimizerType, "Migrad"),
                        Strategy(0),
                        PrintLevel(self.mPrintLevel),
                        Constrain(constrainParams),
                        Offset(RooStats.IsNLLOffset()),
                    )

                    print(f"rebuild using fitted parameter value for B-model snapshot")
                    constrainParams.Print("v")

                    RooStats.SetAllConstant(poiModel, False)

            print(f"StandardHypoTestInvDemo: Initial parameters used for rebuilding: ")
            RooStats.PrintListContent(allParams, ROOT.std.cout)

            tw.Start()
            limDist = calc.GetUpperLimitDistribution(True, self.mNToyToRebuild)
            print(f"Time to rebuild distributions :")
            tw.Print()

            if limDist:
                print(f"Expected limits after rebuild distribution ")
                print(f"expected upper limit  (median of limit distribution) ", limDist.InverseCDF(0.5))
                print(
                    f"expected -1 sig limit (0.16% quantile of limit dist) ",
                    limDist.InverseCDF(ROOT.Math.normal_cdf(-1)),
                )
                print(
                    f"expected +1 sig limit (0.84% quantile of limit dist) ",
                    limDist.InverseCDF(ROOT.Math.normal_cdf(1)),
                )
                print(
                    f"expected -2 sig limit (.025% quantile of limit dist) ",
                    limDist.InverseCDF(ROOT.Math.normal_cdf(-2)),
                )
                print(
                    f"expected +2 sig limit (.975% quantile of limit dist) ",
                    limDist.InverseCDF(ROOT.Math.normal_cdf(2)),
                )

                # Plot the upper limit distribution
                limPlot = SamplingDistPlot(50 if (self.mNToyToRebuild < 200) else 100)
                limPlot.AddSamplingDistribution(limDist)
                limPlot.GetTH1F().SetStats(True)  # display statistics
                limPlot.SetLineColor(kBlue)
                c1 = TCanvas("limPlot", "Upper Limit Distribution")
                limPlot.Draw()
                c1.Update()
                c1.Draw()
                c1.SaveAs("StandardHypoTestInvDemo.1.png")

                # save result in a file
                limDist.SetName("RULDist")
                fileOut = TFile("RULDist.root", "RECREATE")
                limDist.Write()
                fileOut.Close()

                # update r to a new updated result object containing the rebuilt expected p-values distributions
                # (it will not recompute the expected limit)
                if r:
                    del r  # need to delete previous object since GetInterval will return a cloned copy
                r = calc.GetInterval()

            else:
                print(f"ERROR : failed to re-build distributions ")

        return r

    ROOT.RooStats.HypoTestInvTool.RunInverter = RunInverter


# end class HypoTestInvTool
# Loading new definitions of HypoTestInvTool_plus into HypoTestInvTool ...
ROOT.RooStats.HypoTestInvTool = HypoTestInvTool_plus
# Could seem redundant but it would help you if you want expand new-new definitions.


# --------------------------------------------------
def StandardHypoTestInvDemo(
    infile=0,
    wsName="combined",
    modelSBName="ModelConfig",
    modelBName="",
    dataName="obsData",
    calculatorType=0,
    testStatType=0,
    useCLs=True,
    npoints=6,
    poimin=0,
    poimax=5,
    ntoys=1000,
    useNumberCounting=False,
    nuisPriorName=0,
):
    """

    Other Parameter to pass in tutorial
    apart from standard for filename, ws, modelconfig and data

    type = 0 Freq calculator
    type = 1 Hybrid calculator
    type = 2 Asymptotic calculator
    type = 3 Asymptotic calculator using nominal Asimov data sets (not using fitted parameter values but nominal ones)

    testStatType = 0 LEP
    Tevatron
    Likelihood
    mu_hat)
    mu_hat)
    statistic
    statistic

    useCLs          scan for CLs (otherwise for CLs+b)

    npoints = -1

    poimin,poimax:  min/max value to scan in case of fixed scans
    (if min >  max, try to find automatically)

    ntoys:         number of toys to use

    useNumberCounting:  set to True when using number counting events

    nuisPriorName:   name of prior for the nuisance. This is often expressed as constraint term in the global model
    HybridCalculator (type=1)
    If not given by default the prior pdf from ModelConfig is used.

    extra options are available as global parameters of the macro. They major ones are:

    plotHypoTestResult   plot result of tests at each point (TS distributions) (default is True)
    writeResult          write result of scan (default is True)
    rebuild              rebuild scan for expected limits (require extra toys) (default is False)
    generateBinned       generate binned data sets for toys (default is False) - be careful not to activate with
    large (>=3) number of observables
    nToyRatio            ratio of S+B/B toys (default is 2)
    """

    filename = ROOT.TString(infile)
    if filename.IsNull():
        filename = "results/example_combined_GaussExample_model.root"
        fileExist = not ROOT.gSystem.AccessPathName(filename)  # note opposite return code
        # if file does not exists generate with histfactory
        if not fileExist:
            if os.name == "nt":
                print("HistFactory file cannto be generated on Windows - exit")
                return
            # Normally this would be run on the command line
            print(f"will run standard hist2workspace example")
            ROOT.gROOT.ProcessLine(".! prepareHistFactory .")
            ROOT.gROOT.ProcessLine(".! hist2workspace config/example.xml")
            print(f"\n\n---------------------")
            print(f"Done creating example input")
            print(f"---------------------\n\n")

    else:
        filename = infile

    # Try to open the file
    file = ROOT.TFile.Open(filename)

    # if input file was specified but not found, quit
    if not file:
        print(f"StandardRooStatsDemoMacro: Input file {filename} is not found")
        return

    calc = ROOT.RooStats.HypoTestInvTool()  # instance of HypoTestInvTool
    # calc = HypoTestInvTool_plus()           # instance of HypoTestInvTool

    # set parameters
    calc.SetParameter("PlotHypoTestResult", optHTInv.plotHypoTestResult)
    calc.SetParameter("WriteResult", optHTInv.writeResult)
    calc.SetParameter("Optimize", optHTInv.optimize)
    calc.SetParameter("UseVectorStore", optHTInv.useVectorStore)
    calc.SetParameter("GenerateBinned", optHTInv.generateBinned)
    calc.SetParameter("NToysRatio", optHTInv.nToysRatio)
    calc.SetParameter("MaxPOI", optHTInv.maxPOI)
    calc.SetParameter("EnableDetailedOutput", optHTInv.enableDetailedOutput)
    calc.SetParameter("Rebuild", optHTInv.rebuild)
    calc.SetParameter("ReuseAltToys", optHTInv.reuseAltToys)
    calc.SetParameter("NToyToRebuild", optHTInv.nToyToRebuild)
    calc.SetParameter("RebuildParamValues", optHTInv.rebuildParamValues)
    calc.SetParameter("MassValue", optHTInv.massValue)
    calc.SetParameter("MinimizerType", optHTInv.minimizerType)
    calc.SetParameter("PrintLevel", optHTInv.printLevel)
    calc.SetParameter("InitialFit", optHTInv.initialFit)
    calc.SetParameter("ResultFileName", optHTInv.resultFileName)
    calc.SetParameter("RandomSeed", optHTInv.randomSeed)
    calc.SetParameter("AsimovBins", optHTInv.nAsimovBins)

    # enable offset for all roostats
    if optHTInv.useNLLOffset:
        RooStats.SetNLLOffsetMode("initial")

    w = file.Get(wsName)
    r = 0
    print(w, "\t", filename)
    if w != ROOT.kNone:
        r = calc.RunInverter(
            w,
            modelSBName,
            modelBName,
            dataName,
            calculatorType,
            testStatType,
            useCLs,
            npoints,
            poimin,
            poimax,
            ntoys,
            useNumberCounting,
            nuisPriorName,
        )
        if not r:
            raise RuntimeError("Error running the HypoTestInverter - Exit ")
            return

    else:
        # case workspace is not present look for the inverter result
        print(f"Reading an HypoTestInverterResult with name {wsName} from file " << filename)
        r = file.Get(wsName)  # dynamic_cast
        if not r:
            raise RuntimeError("File {filename} does not contain a workspace or an HypoTestInverterResult - Exit ")

            file.ls()
            return

    calc.AnalyzeResult(r, calculatorType, testStatType, useCLs, npoints, infile)

    return


# --------------------------------------------------


def ReadResult(fileName, resultName="", useCLs=True):
    # read a previous stored result from a file given the result name
    StandardHypoTestInvDemo(fileName, resultName, "", "", "", 0, 0, useCLs)


# ifdef USE_AS_MAIN
def main():
    StandardHypoTestInvDemo()


main()
# endif



## \file
## \ingroup tutorial_roofit_main
## \notebook -nodraw
## Organization and simultaneous fits: using RooSimWSTool to construct a simultaneous pdf
## that is built of variations of an input pdf
##
## \macro_code
## \macro_output
##
## \date February 2018
## \authors Clemens Lange, Wouter Verkerke (C++ version)

import ROOT


# Create master pdf
# ---------------------------------

# Construct gauss(x,m,s)
x = ROOT.RooRealVar("x", "x", -10, 10)
m = ROOT.RooRealVar("m", "m", 0, -10, 10)
s = ROOT.RooRealVar("s", "s", 1, -10, 10)
gauss = ROOT.RooGaussian("g", "g", x, m, s)

# Construct poly(x,p0)
p0 = ROOT.RooRealVar("p0", "p0", 0.01, 0.0, 1.0)
poly = ROOT.RooPolynomial("p", "p", x, [p0])

# model = f*gauss(x) + (1-f)*poly(x)
f = ROOT.RooRealVar("f", "f", 0.5, 0.0, 1.0)
model = ROOT.RooAddPdf("model", "model", [gauss, poly], [f])

# Create category observables for splitting
# ----------------------------------------------------------------------------------

# Define two categories that can be used for splitting
c = ROOT.RooCategory("c", "c")
c.defineType("run1")
c.defineType("run2")

d = ROOT.RooCategory("d", "d")
d.defineType("foo")
d.defineType("bar")

# Set up SimWSTool
# -----------------------------

# Import ingredients in a workspace
w = ROOT.RooWorkspace("w", "w")
w.Import({model, c, d})

# Make Sim builder tool
sct = ROOT.RooSimWSTool(w)

# Build a simultaneous model with one split
# ---------------------------------------------------------------------------------

# Construct a simultaneous pdf with the following form
#
# model_run1(x) = f*gauss_run1(x,m_run1,s) + (1-f)*poly
# model_run2(x) = f*gauss_run2(x,m_run2,s) + (1-f)*poly
# simpdf(x,c) = model_run1(x) if c=="run1"
#             = model_run2(x) if c=="run2"
#
# Returned pdf is owned by the workspace
model_sim = sct.build("model_sim", "model", SplitParam=("m", "c"))

# Print tree structure of model
model_sim.Print("t")

# Adjust model_sim parameters in workspace
w.var("m_run1").setVal(-3)
w.var("m_run2").setVal(+3)

# Print contents of workspace
w.Print("v")

# Build a simultaneous model with product split
# -----------------------------------------------------------------------------------------

# Build another simultaneous pdf using a composite split in states c X d
model_sim2 = sct.build("model_sim2", "model", SplitParam=("p0", "c,d"))

# Print tree structure of self model
model_sim2.Print("t")



## \file
## \ingroup tutorial_roofit_main
## \notebook -nodraw
## Organization and simultaneous fits: illustration use of ROOT.RooCustomizer and
## ROOT.RooSimWSTool interface in factory workspace tool in a complex standalone B physics example
##
## \macro_code
## \macro_output
##
## \date February 2018
## \authors Clemens Lange, Wouter Verkerke (C++ version)

import ROOT


w = ROOT.RooWorkspace("w")

# Build a complex example pdf
# -----------------------------------------------------------

# Make signal model for CPV: A bmixing decay function in t (convoluted with a triple Gaussian resolution model)
# times a Gaussian function the reconstructed mass
w.factory(
    "PROD::sig(  BMixDecay::sig_t( dt[-20,20], mixState[mixed=1,unmix=-1], tagFlav[B0=1,B0bar=-1], "
    "tau[1.54], dm[0.472], w[0.05], dw[0], "
    "AddModel::gm({GaussModel(dt,biasC[-10,10],sigmaC[0.1,3],dterr[0.01,0.2]), "
    "GaussModel(dt,0,sigmaT[3,10]), "
    "GaussModel(dt,0,20)},{fracC[0,1],fracT[0,1]}), "
    "DoubleSided ), "
    "Gaussian::sig_m( mes[5.20,5.30], mB0[5.20,5.30], sigmB0[0.01,0.05] ))"
)

# Make background component: A plain decay function in t times an Argus
# function in the reconstructed mass
w.factory("PROD::bkg(  Decay::bkg_t( dt, tau, gm, DoubleSided), " "ArgusBG::bkg_m( mes, 5.291, k[-100,-10]))")

# Make composite model from the signal and background component
w.factory("SUM::model( Nsig[5000,0,10000]*sig, NBkg[500,0,10000]*bkg )")

# Example of RooSimWSTool interface
# ------------------------------------------------------------------

# Introduce a flavour tagging category tagCat as observable with 4 states corresponding
# to 4 flavour tagging techniques with different performance that require different
# parameterizations of the fit model
#
# ROOT.RooSimWSTool operation:
#     - Make 4 clones of model (for each tagCat) state, will gain an individual
#       copy of parameters w, and biasC. The other parameters remain common
#     - Make a simultaneous pdf of the 4 clones assigning each to the appropriate
#       state of the tagCat index category

# ROOT.RooSimWSTool is interfaced as meta-type SIMCLONE in the factory. The $SplitParam()
# argument maps to the SplitParam() named argument in the
# ROOT.RooSimWSTool constructor
w.factory("SIMCLONE::model_sim( model, $SplitParam({w,dw,biasC},tagCat[Lep,Kao,NT1,NT2]))")

# Example of RooCustomizer interface
# -------------------------------------------------------------------
#
# Class ROOT.RooCustomizer makes clones of existing pdfs with certain prescribed
# modifications (branch of leaf node replacements)
#
# Here we take our model (the original before ROOT.RooSimWSTool modifications)
# and request that the parameter w (the mistag rate) is replaced with
# an expression-based function that calculates w in terms of the Dilution
# parameter D that is defined D = 1-2*w

# Make a clone model_D of original 'model' replacing 'w' with
# 'expr('0.5-D/2',D[0,1])'
w.factory("EDIT::model_D(model, w=expr('0.5-D/2',D[0,1]) )")

# Print workspace contents
w.Print()

# Make workspace visible on command line
ROOT.gDirectory.Add(w)



## \file
## \ingroup tutorial_ml
## \notebook
##
## \macro_image
## \macro_output
## \macro_code
##
## \author Harshal Shende

import torch
from torch import nn

# Define model

print("running Torch code defining the model....")

# Custom Reshape Layer
class Reshape(torch.nn.Module):
    def forward(self, x):
        return x.view(-1,1,16,16)

# CNN Model Definition
net = torch.nn.Sequential(
    Reshape(),
    nn.Conv2d(1, 10, kernel_size=3, padding=1),
    nn.ReLU(),
    nn.BatchNorm2d(10),
    nn.Conv2d(10, 10, kernel_size=3, padding=1),
    nn.ReLU(),
    nn.MaxPool2d(kernel_size=2),
    nn.Flatten(),
    nn.Linear(10*8*8, 256),
    nn.ReLU(),
    nn.Linear(256, 2),
    nn.Sigmoid()
    )

# Construct loss function and Optimizer.
criterion = nn.BCELoss()
optimizer = torch.optim.Adam


def fit(model, train_loader, val_loader, num_epochs, batch_size, optimizer, criterion, save_best, scheduler):
    trainer = optimizer(model.parameters(), lr=0.01)
    schedule, schedulerSteps = scheduler
    best_val = None

    # Setup GPU
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    model = model.to(device)

    for epoch in range(num_epochs):
        # Training Loop
        # Set to train mode
        model.train()
        running_train_loss = 0.0
        running_val_loss = 0.0
        for i, (X, y) in enumerate(train_loader):
            trainer.zero_grad()
            X, y = X.to(device), y.to(device)
            output = model(X)
            target = y
            train_loss = criterion(output, target)
            train_loss.backward()
            trainer.step()

            # print train statistics
            running_train_loss += train_loss.item()
            if i % 4 == 3:    # print every 4 mini-batches
                print(f"[{epoch+1}, {i+1}] train loss: {running_train_loss / 4 :.3f}")
                running_train_loss = 0.0

        if schedule:
            schedule(optimizer, epoch, schedulerSteps)

        # Validation Loop
        # Set to eval mode
        model.eval()
        with torch.no_grad():
            for i, (X, y) in enumerate(val_loader):
                X, y = X.to(device), y.to(device)
                output = model(X)
                target = y
                val_loss = criterion(output, target)
                running_val_loss += val_loss.item()

            curr_val = running_val_loss / len(val_loader)
            if save_best:
               if best_val==None:
                   best_val = curr_val
               best_val = save_best(model, curr_val, best_val)

            # print val statistics per epoch
            print(f"[{epoch+1}] val loss: {curr_val :.3f}")
            running_val_loss = 0.0

    print(f"Finished Training on {epoch+1} Epochs!")

    return model


def predict(model, test_X, batch_size=100):
    # Set to eval mode

    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    model = model.to(device)

    model.eval()


    test_dataset = torch.utils.data.TensorDataset(torch.Tensor(test_X))
    test_loader = torch.utils.data.DataLoader(test_dataset, batch_size=batch_size, shuffle=False)

    predictions = []
    with torch.no_grad():
        for i, data in enumerate(test_loader):
            X = data[0].to(device)
            outputs = model(X)
            predictions.append(outputs)
        preds = torch.cat(predictions)

    return preds.cpu().numpy()


load_model_custom_objects = {"optimizer": optimizer, "criterion": criterion, "train_func": fit, "predict_func": predict}

# Store model to file
m = torch.jit.script(net)
torch.jit.save(m,"PyTorchModelCNN.pt")
print("The PyTorch CNN model is created and saved as PyTorchModelCNN.pt") 



# -----------------------------------------------------------------------------
#  Author: Danilo Piparo <Danilo.Piparo@cern.ch> CERN
#  Author: Enric Tejedor <enric.tejedor.saavedra@cern.ch> CERN
# -----------------------------------------------------------------------------

################################################################################
# Copyright (C) 1995-2020, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

from IPython.core.magic import Magics, cell_magic, magics_class
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring

from ROOT._jupyroot.helpers import utils


@magics_class
class CppMagics(Magics):
    @cell_magic
    @magic_arguments()
    @argument("-a", "--aclic", action="store_true", help="Compile code with ACLiC.")
    @argument("-d", "--declare", action="store_true", help="Declare functions and/or classes.")
    def cpp(self, line, cell):
        """Executes the content of the cell as C++ code."""
        args = parse_argstring(self.cpp, line)
        if args.aclic:
            utils.invokeAclic(cell)
        elif args.declare:
            utils.declareCppCode(cell)
        else:
            utils.processMagicCppCode(cell)


def load_ipython_extension(ipython):
    ipython.register_magics(CppMagics)



# Author: Vincenzo Eduardo Padulano 12/2024

################################################################################
# Copyright (C) 1995-2024, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

from . import pythonization
from ._memory_utils import inject_constructor_releasing_ownership
from ._uhi import _add_plotting_features, _add_serialization_features


# Fill with array-like data
def _FillWithArrayTH2(self, *args):
    """
    Fill a histogram using array-like input.
    Parameters:
    - self: histogram
    - args: arguments to FillN
            If the first 2 arguments are array-like:
            - converts them to numpy arrays
            - fills the histogram with these arrays
            - optional third argument is weights array,
              if not provided, weights of 1 are used
            Otherwise:
            - Arguments are passed directly to the original FillN method
    Returns:
    - Result of FillN if array case is detected, otherwise result of Fill
    Raises:
    - ValueError: If x, y, and/or weights do not have matching lengths
    """
    # If there are less than 2 arguments, cannot do vectorized Fill
    if len(args) < 2:
        return self._Fill(*args)

    import numpy as np

    try:
        x = np.asanyarray(args[0], dtype=np.float64)
        y = np.asanyarray(args[1], dtype=np.float64)

        if len(x) != len(y):
            raise ValueError(f"Length mismatch: x length ({len(x)}) != y length ({len(y)})")

        n = len(x)
    except Exception:
        # Not convertible
        return self._Fill(*args)

    if len(args) >= 3 and args[2] is not None:
        weights = np.asanyarray(args[2], dtype=np.float64)
        if len(weights) != n:
            raise ValueError(f"Length mismatch: data length ({n}) != weights length ({len(weights)})")
    else:
        weights = np.ones(n)

    return self.FillN(n, x, y, weights)


def _TH2Poly_AddBin(self, *args, **kwargs):
    """
    The TH2Poly always takes ownership of the added bin objects.
    """
    from ROOT._pythonization._memory_utils import declare_cpp_owned_arg

    declare_cpp_owned_arg(0, "poly", args, kwargs)

    self._AddBin(*args, **kwargs)


# The constructors need to be pythonized for each derived class separately:
_th2_derived_classes_to_pythonize = [
    "TH2C",
    "TH2S",
    "TH2I",
    "TH2L",
    "TH2F",
    "TH2D",
    # "TH2Poly", # Derives from TH2 but does not automatically register
    # "TH2PolyBin", Does not derive from TH2
    "TProfile2D",
    # "TProfile2PolyBin", Derives from TH2PolyBin which does not derive from TH2
    "TProfile2Poly",
]

for klass in _th2_derived_classes_to_pythonize:
    pythonization(klass)(inject_constructor_releasing_ownership)

    # Add UHI plotting features
    pythonization(klass)(_add_plotting_features)

    # Add serialization features
    pythonization(klass)(_add_serialization_features)

    # Support vectorized Fill
    @pythonization(klass)
    def _enable_numpy_fill(klass):
        klass._Fill = klass.Fill
        klass.Fill = _FillWithArrayTH2


@pythonization("TH2Poly")
def _pythonize_th2poly(klass):
    klass._AddBin = klass.AddBin
    klass.AddBin = _TH2Poly_AddBin



## \file
## \ingroup tutorial_graphs
## \notebook
## \preview A macro to demonstrate the functionality of TGraph::Apply() method.
## TGraph::Apply applies a function `f` to all the data TGraph points, `f` may be a 1-D function TF1 or 2-d function TF2.
## The Y values of the graph are replaced by the ROOT.values computed using the function.
##
##
## The Apply() method can be used as well for TGraphErrors and TGraphAsymmErrors.
##
## \macro_image
## \macro_code
##
## \author Miro Helbich, Jamie Gooding

import numpy as np
import ROOT

npoints = 3
xaxis = np.array([1.0, 2.0, 3.0])
yaxis = np.array([10.0, 20.0, 30.0])

gr1 = ROOT.TGraph(npoints, xaxis, yaxis)
ff = ROOT.TF2("ff", "-1./y")  # Defining the function `f`

c1 = ROOT.TCanvas("c1", "c1", 0, 0, 700, 500)
c1.Divide(2, 1)

c1.cd(1)
gr1.DrawClone("A*")  # Using DrawClone to create a copy of the graph in the canvas.
c1.cd(2)
gr1.Apply(ff)  # Applies the function `f` to all the data TGraph points
gr1.Draw("A*")
# Without DrawClone, the modifications to gr1 via Apply(ff) are reflected in the original graph
# displayed in c1 (the two drawn graphs are not independent).



## \file
## \ingroup tutorial_roofit_main
## \notebook
## Special pdf's: using non-parametric (multi-dimensional) kernel estimation pdfs
##
## \macro_image
## \macro_code
## \macro_output
##
## \date February 2018
## \authors Clemens Lange, Wouter Verkerke (C++ version)

import ROOT


# Create low stats 1D dataset
# -------------------------------------------------------

# Create a toy pdf for sampling
x = ROOT.RooRealVar("x", "x", 0, 20)
p = ROOT.RooPolynomial("p", "p", x, [0.01, -0.01, 0.0004])

# Sample 500 events from p
data1 = p.generate({x}, 200)

# Create 1D kernel estimation pdf
# ---------------------------------------------------------------

# Create adaptive kernel estimation pdf. In self configuration the input data
# is mirrored over the boundaries to minimize edge effects in distribution
# that do not fall to zero towards the edges
kest1 = ROOT.RooKeysPdf("kest1", "kest1", x, data1, ROOT.RooKeysPdf.MirrorBoth)

# An adaptive kernel estimation pdf on the same data without mirroring option
# for comparison
kest2 = ROOT.RooKeysPdf("kest2", "kest2", x, data1, ROOT.RooKeysPdf.NoMirror)

# Adaptive kernel estimation pdf with increased bandwidth scale factor
# (promotes smoothness over detail preservation)
kest3 = ROOT.RooKeysPdf("kest1", "kest1", x, data1, ROOT.RooKeysPdf.MirrorBoth, 2)

# Plot kernel estimation pdfs with and without mirroring over data
frame = x.frame(Title="Adaptive kernel estimation pdf with and w/o mirroring", Bins=20)
data1.plotOn(frame)
kest1.plotOn(frame)
kest2.plotOn(frame, LineStyle="--", LineColor="r")

# Plot kernel estimation pdfs with regular and increased bandwidth
frame2 = x.frame(Title="Adaptive kernel estimation pdf with regular, bandwidth")
kest1.plotOn(frame2)
kest3.plotOn(frame2, LineColor="m")

# Create low status 2D dataset
# -------------------------------------------------------

# Construct a 2D toy pdf for sampleing
y = ROOT.RooRealVar("y", "y", 0, 20)
py = ROOT.RooPolynomial(
    "py",
    "py",
    y,
    [0.01, 0.01, -0.0004],
)
pxy = ROOT.RooProdPdf("pxy", "pxy", [p, py])
data2 = pxy.generate({x, y}, 1000)

# Create 2D kernel estimation pdf
# ---------------------------------------------------------------

# Create 2D adaptive kernel estimation pdf with mirroring
kest4 = ROOT.RooNDKeysPdf("kest4", "kest4", [x, y], data2, "am")

# Create 2D adaptive kernel estimation pdf with mirroring and double
# bandwidth
kest5 = ROOT.RooNDKeysPdf("kest5", "kest5", [x, y], data2, "am", 2)

# Create a histogram of the data
hh_data = data2.createHistogram("hh_data", x, Binning=10, YVar=dict(var=y, Binning=10))

# Create histogram of the 2d kernel estimation pdfs
hh_pdf = kest4.createHistogram("hh_pdf", x, Binning=25, YVar=dict(var=y, Binning=25))
hh_pdf2 = kest5.createHistogram("hh_pdf2", x, Binning=25, YVar=dict(var=y, Binning=25))
hh_pdf.SetLineColor("kBlue")
hh_pdf2.SetLineColor("kMagenta")

c = ROOT.TCanvas("rf707_kernelestimation", "rf707_kernelestimation", 800, 800)
c.Divide(2, 2)
c.cd(1)
ROOT.gPad.SetLeftMargin(0.15)
frame.GetYaxis().SetTitleOffset(1.4)
frame.Draw()
c.cd(2)
ROOT.gPad.SetLeftMargin(0.15)
frame2.GetYaxis().SetTitleOffset(1.8)
frame2.Draw()
c.cd(3)
ROOT.gPad.SetLeftMargin(0.15)
hh_data.GetZaxis().SetTitleOffset(1.4)
hh_data.Draw("lego")
c.cd(4)
ROOT.gPad.SetLeftMargin(0.20)
hh_pdf.GetZaxis().SetTitleOffset(2.4)
hh_pdf.Draw("surf")
hh_pdf2.Draw("surfsame")

c.SaveAs("rf707_kernelestimation.png")



## \file
## \ingroup tutorial_sql
##
## Query example to MySQL test database.
## Example of query by using the test database made in MySQL, you need the
## database test installed in localhost, with user nobody without password.
##
## Based on sqlselect.C by Sergey Linev
##
## \macro_code
##
## \author Juan Fernando Jaramillo Botero

from ROOT import TSQLServer, TSQLResult, TSQLRow, TStopwatch


db = TSQLServer.Connect("mysql://localhost/test", "nobody", "")

print("Server info: %s" % db.ServerInfo())

# list databases available on server
print("")
print("List all databases on server %s" % db.GetHost())
res = db.GetDataBases()
row = res.Next()
while row:
    print("%s" % row.GetField(0))
    row = res.Next()

# list tables in database "test" (the permission tables)
print('')
print('List all tables in database "test" on server %s' % db.GetHost())
res = db.GetTables("test")
row = res.Next()
while row:
    print("%s" % row.GetField(0))
    row = res.Next()

# list columns in table "runcatalog" in database "mysql"
print('')
print('List all columns in table "runcatalog" in database "test" on server %s' %
      db.GetHost())
res = db.GetColumns("test", "runcatalog")
row = res.Next()
while row:
    print("%s" % row.GetField(0))
    row = res.Next()

# start timer
timer = TStopwatch()
timer.Start()

# query database and print results
# sql = "select dataset,rawfilepath from test.runcatalog " \
#       "WHERE tag&(1<<2) AND (run=490001 OR run=300122)"
sql = "select count(*) from test.runcatalog " \
      "WHERE tag&(1<<2)"

res = db.Query(sql)

nrows = res.GetRowCount()
print("")
print("Got %d rows in result" % nrows)

nfields = res.GetFieldCount()
for i in range(nfields):
    print("%40s" % res.GetFieldName(i))
print("")
print("=" * (nfields * 40))
print("")

for i in range(nrows):
    row = res.Next()
    for j in range(nfields):
        print("%40s" % row.GetField(j))
    print("")

# stop timer and print results
timer.Stop()
rtime = timer.RealTime()
ctime = timer.CpuTime()

print("")
print("RealTime=%f seconds, CpuTime=%f seconds" % (rtime, ctime))



## \file
## \ingroup tutorial_sql
## \notebook -nodraw
## Fill run catalog with nfiles entries
##
## Based on sqlfill.C by Sergey Linev
##
## \macro_code
##
## \author Juan Fernando Jaramillo Botero

from ROOT import TSQLServer, TSQLResult, TStopwatch, gRandom


nfiles = 1000

ins = "INSERT INTO runcatalog VALUES ('%s', %d," \
      " %d, %d, %d, %10.2f, '%s', '%s', '1997-01-15 20:16:28'," \
      " '1999-01-15 20:16:28', '%s', '%s')"

tag = evt = 0

# open connection to MySQL server on localhost
db = TSQLServer.Connect("mysql://localhost/test", "nobody", "")

# first clean table of old entries
res = db.Query("DELETE FROM runcatalog")

# start timer
timer = TStopwatch()
timer.Start()

# fill run catalog
for i in range(nfiles):
    dataset = "testrun_%d" % i
    rawfile = "/v1/data/lead/test/run_%d.root" % i
    tag = int(gRandom.Rndm() * 10.)
    sql = ins % (dataset, i, evt, evt + 10000, tag, 25.5, "test", "lead",
                 rawfile, "test run dummy data")
    evt += 10000
    res = db.Query(sql)
    # print("%s" % sql)


# stop timer and print results
timer.Stop()
rtime = timer.RealTime()
ctime = timer.CpuTime()

print("")
print("%d files in run catalog" % nfiles)
print("RealTime=%f seconds, CpuTime=%f seconds" % (rtime, ctime))



def MakeKerasDense(layer):
    """
    Create a Keras-compatible dense (fully connected) layer operation using SOFIE framework.

    This function takes a dictionary representing a dense layer and its attributes and
    constructs a Keras-compatible dense (fully connected) layer operation using the SOFIE framework.
    A dense layer applies a matrix multiplication between the input tensor and weight matrix,
    and adds a bias term.

    Parameters:
    layer (dict): A dictionary containing layer information including input, output,
                  layer weight names, and data type - must be float.

    Returns:
    ROperator_Gemm: A SOFIE framework operator representing the dense layer operation.
    """
    from ROOT.TMVA.Experimental import SOFIE

    finput = layer["layerInput"]
    foutput = layer["layerOutput"]
    fLayerDType = layer["layerDType"]
    fLayerInputName = finput[0]
    fLayerOutputName = foutput[0]
    fWeightNames = layer["layerWeight"]
    fKernelName = fWeightNames[0]
    fBiasName = fWeightNames[1]
    attr_alpha = 1.0
    attr_beta = 1.0
    attr_transA = 0
    attr_transB = 0
    if SOFIE.ConvertStringToType(fLayerDType) == SOFIE.ETensorType.FLOAT:
        op = SOFIE.ROperator_Gemm["float"](
            attr_alpha, attr_beta, attr_transA, attr_transB, fLayerInputName, fKernelName, fBiasName, fLayerOutputName
        )
        return op
    else:
        raise RuntimeError("TMVA::SOFIE - Unsupported - Operator Gemm does not yet support input type " + fLayerDType)



## \file
## \ingroup tutorial_hist
## \notebook
## Fill a 1D histogram with random values using predefined functions.
##
## \macro_image
## \macro_code
##
## \date November 2024
## \author Giacomo Parolini
import ROOT

# Create a one dimensional histogram and fill it with a gaussian distribution
h1d = ROOT.TH1D("h1d", "Test random numbers", nbinsx = 200, xlow = 0.0, xup = 10.0)

# "gaus" is a predefined ROOT function. Here we are filling the histogram with
# 10000 values sampled from that distribution.
h1d.FillRandom("gaus", 10000)

# Open a ROOT file and save the histogram
with ROOT.TFile.Open("fillrandom_py.root", "RECREATE") as myfile:
   myfile.WriteObject(h1d, h1d.GetName())



# -*- coding:utf-8 -*-
# -----------------------------------------------------------------------------
#  Copyright (c) 2016, ROOT Team.
#  Authors: Danilo Piparo <Danilo.Piparo@cern.ch> CERN
# -----------------------------------------------------------------------------

################################################################################
# Copyright (C) 1995-2020, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

from IPython.core.magic import Magics, line_magic, magics_class
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring

from ROOT._jupyroot.helpers.utils import enableJSVis


@magics_class
class JSRootMagics(Magics):
    def __init__(self, shell):
        super(JSRootMagics, self).__init__(shell)

    @line_magic
    @magic_arguments()
    @argument(
        "arg",
        nargs="?",
        default="on",
        help="Enable or disable JavaScript visualisation. Possible values: on (default), off",
    )
    def jsroot(self, line):
        """Change the visualisation of plots from images to interactive JavaScript objects."""
        args = parse_argstring(self.jsroot, line)
        if args.arg == "on":
            enableJSVis(True)
        elif args.arg == "off":
            enableJSVis(False)


def load_ipython_extension(ipython):
    ipython.register_magics(JSRootMagics)



## \file
## \ingroup Tutorials
## To run, do "python <path-to>/demos.py"
##
## \macro_code
##
## \author Wim Lavrijsen, Enric Tejedor

import os, sys
import ROOT

# To run, do "python <path-to>/demos.py"

# enable running from another directory than the one where demos.py resides
workdir = os.path.dirname( sys.argv[0] )
if workdir:
   os.chdir( workdir )

# This macro generates a Controlbar menu.
# To execute an item, click with the left mouse button.
# To see the HELP of a button, click on the right mouse button.

ROOT.gStyle.SetScreenFactor(1)   # if you have a large screen, select 1.2 or 1.4

bar = ROOT.TControlBar( 'vertical', 'Demos', 10, 10 )

# The callbacks to python work by having CLING call the python interpreter through
# the "TPython" class. Note the use of "raw strings."
to_run = 'exec(open(\'{}\').read())'


bar.AddButton( 'Help on Demos', r'TPython::Exec( "' + to_run.format('demoshelp.py') + '" );', 'Click Here For Help on Running the Demos' )
bar.AddButton( 'browser',       r'TPython::Exec( "b = ROOT.TBrowser()" );',          'Start the ROOT browser' )
bar.AddButton( 'hsimple',       r'TPython::Exec( "' + to_run.format('hsimple.py') + '" );',   'Creating histograms/Ntuples on file', "button" )
bar.AddButton( 'hsum',          r'TPython::Exec( "' + to_run.format('hist/hsum.py') + '" );',      'Filling Histograms and Some Graphics Options' )
bar.AddButton( 'formula1',      r'TPython::Exec( "' + to_run.format('visualisation/graphics/formula1.py') + '" );',  'Simple Formula and Functions' )
bar.AddButton( 'surfaces',      r'TPython::Exec( "' + to_run.format('visualisation/graphics/surfaces.py') + '" );',  'Surface Drawing Options' )
bar.AddButton( 'fillrandom',    r'TPython::Exec( "' + to_run.format('hist/fillrandom.py') + '" );','Histograms with Random Numbers from a Function' )
bar.AddButton( 'fit1',          r'TPython::Exec( "' + to_run.format('math/fit/fit1.py') + '" );',      'A Simple Fitting Example' )
bar.AddButton( 'multifit',      r'TPython::Exec( "' + to_run.format('math/fit/multifit.py') + '" );',  'Fitting in Subranges of Histograms' )
bar.AddButton( 'h1draw',        r'TPython::Exec( "' + to_run.format('hist/hist015_TH1_read_and_draw.py') + '" );',    'Drawing Options for 1D Histograms' )
bar.AddButton( 'graph',         r'TPython::Exec( "' + to_run.format('visualisation/graphs/graph.py') + '" );',     'Example of a Simple Graph' )
bar.AddButton( 'gerrors',       r'TPython::Exec( "' + to_run.format('visualisation/graphs/gerrors.py') + '" );',   'Example of a Graph with Error Bars' )
bar.AddButton( 'tornado',       r'TPython::Exec( "' + to_run.format('visualisation/graphics/tornado.py') + '" );',   'Examples of 3-D PolyMarkers' )
bar.AddButton( 'shapes',        r'TPython::Exec( "' + to_run.format('legacy/g3d/shapes.py') + '" );',    'The Geometry Shapes' )
bar.AddButton( 'geometry',      r'TPython::Exec( "' + to_run.format('legacy/g3d/geometry.py') + '" );',  'Creation of the NA49 Geometry File' )
bar.AddButton( 'na49view',      r'TPython::Exec( "' + to_run.format('legacy/g3d/na49view.py') + '" );',  'Two Views of the NA49 Detector Geometry' )
bar.AddButton( 'ntuple1',       r'TPython::Exec( "' + to_run.format('io/tree/ntuple1.py') + '" );',   'Ntuples and Selections' )
bar.AddSeparator()       # not implemented
bar.AddButton( 'make ntuple',   r'TPython::Exec( "' + to_run.format('io/tree/csv2tntuple.py') + '" );',       'Convert a text file to an ntuple' )

bar.Show()

ROOT.gROOT.SaveContext()


## wait for input to keep the GUI (which lives on a ROOT event dispatcher) alive
if __name__ == '__main__':
   rep = ''
   while not rep in [ 'q', 'Q' ]:
      rep = input( 'enter "q" to quit: ' )
      if 1 < len(rep):
         rep = rep[0]



# Author: Enric Tejedor CERN  02/2019

################################################################################
# Copyright (C) 1995-2019, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

from . import pythonization
from ._generic import _add_getitem_checked


@pythonization('TVector3')
def pythonize_tvector3(klass):
    # Parameters:
    # klass: class to be pythonized

    # Add checked __getitem__.
    # Allows to throw pythonic IndexError when index is out of range
    # and to iterate over the vector.
    _add_getitem_checked(klass)
